# -*- coding: utf-8 -*-
"""Module for a sensitivity analysis for one experiment"""
import os
import six
from functools import partial
from argparse import Namespace
import yaml
import os.path as osp
from collections import namedtuple
from itertools import product, chain, repeat, starmap
import logging
import numpy as np
from gwgen.main import OrderedDict, GWGENOrganizer
import gwgen.utils as utils
from gwgen.utils import isstring, docstrings
import matplotlib as mpl


class SensitivityAnalysis(object):
    """Class that performs and manages a sensitivity analysis for the given
    organizer"""

    @property
    def logger(self):
        return logging.getLogger('%s.%s' % (__name__, self.__class__.__name__))

    @property
    def experiment(self):
        """The base experiment of this sensitivity analysis"""
        return self.organizer.experiment

    @property
    def experiments(self):
        projectname = self.projectname
        return self.organizer.config.experiments.project_map[projectname]

    @property
    def exp_config(self):
        return self.organizer.exp_config

    @property
    def config(self):
        """The configuration of the sensitivity analysis"""
        return self.exp_config.setdefault('sensitivity_analysis',
                                          OrderedDict())

    @property
    def projectname(self):
        """The projectname of the sensitivity analysis"""
        return self.config['project']

    @projectname.setter
    def projectname(self, value):
        self.config['project'] = value

    @property
    def project_config(self):
        """The project configuration of the sensitivity analysis"""
        return self.organizer.config.projects.setdefault(self.projectname,
                                                         OrderedDict())

    @property
    def sa_organizer(self):
        """The organizer for this sensitivity analysis"""
        try:
            return self._sa_organizer
        except AttributeError:
            self._sa_organizer = GWGENOrganizer(self.organizer.config)
            try:
                projectname = self.projectname
            except KeyError:
                raise ValueError(
                    "No setup of the sensitivity analysis has been done yet! "
                    "Please run the setup method!")
            self._sa_organizer.projectname = projectname
        return self._sa_organizer

    def __init__(self, organizer):
        """
        Parameters
        ----------
        organizer: gwgen.main.GWGENOrganizer
            The organizer to perform thre sensitivity analysis for
        """
        self.organizer = organizer

    @docstrings.get_sectionsf('SensitivityAnalysis._parallelize_command')
    @docstrings.dedent
    def _parallelize_command(self, kws, experiments=None,
                             loop_exps=False):
        """
        Run a GWGENOrganizer command in parallel for all the experiments in
        this analysis

        Parameters
        ----------
        kws: dict
            Keys must be the name of a command of the :attr:`organizer` of this
            analysis, values must be dictionaries for the corresponding command
        experiments: list of str
            The experiments to use. If None, all experiments are used
        loop_exps: bool
            If True the experiments of the sensitivity analysis are called
            one after each other. Otherwise they are called in parallel, each
            on a single core (unless the the serial option in the global
            configuration is set)
        """
        import multiprocessing as mp
        experiments = experiments or self.experiments
        if len(experiments) == 1 and osp.exists(experiments[0]):
            with open(experiments[0]) as f:
                experiments = [l.rstrip() for l in f.readlines()]
        config = self.organizer.global_config
        commands = self.organizer.commands + list(
            self.organizer.parser_commands.values())
        if not loop_exps and not config.get('serial'):
            all_kws = (
                {key: dict(chain([('experiment', exp)], kws[key].items()))
                 for key in kws.keys() if key in commands}
                for exp in experiments)
            nprocs = config.get('nprocs', 'all')
            if nprocs == 'all':
                nprocs = mp.cpu_count()
            nprocs = min([nprocs, len(experiments)])
            config['serial'] = True
            self.logger.debug('Starting %i processes', nprocs)
            pool = mp.Pool(nprocs)
            res = pool.map_async(self, all_kws)
            for (organizer, ns), experiment in zip(res.get(), experiments):
                changed = organizer.config.experiments[experiment]
                self.organizer.config.experiments[experiment] = changed
            config['serial'] = False
            pool.close()
            pool.join()
            pool.terminate()
        else:
            for experiment in experiments or self.experiments:
                for key, val in kws.items():
                    if key in commands:
                        val['experiment'] = experiment
                kws['experiment'] = experiment
                self.sa_organizer.start(**kws)

    def __call__(self, kws):
        """Call the :meth:`~model_organization.ModelOrganizer.start` of the
        :attr:`organizer` attribute

        Parameters
        ----------
        kws: dict
            Any keywords that are passed to the
            :meth:`~model_organization.ModelOrganizer.start` method

        Returns
        -------
        model_organization.ModelOrganizer
            The :attr:`sa_organizer`
        argparse.Namespace
            The return of the :meth:`~model_organization.ModelOrganizer.start`
            method
        """
        return self.sa_organizer, self.sa_organizer.start(**kws)

    def setup(self, root_dir=None, projectname=None, link=True,
              no_move=False, *args, **kwargs):
        """
        Set up the experiments for the project

        Parameters
        ----------
        root_dir: str
            The path to the root directory where the experiments, etc. will
            be stored. If not given, a new directory will be created at the
            same location as the original project
        projectname: str
            The name of the project that shall be initialized at `root_dir`. A
            new directory will be created namely
            ``root_dir + '/' + projectname``. If not given, defaults to
            ``'sensitivity_<id>'`` where ``'<id>'`` is the experiment id
        link: bool
            If set, the source files are linked to the original ones instead
            of copied
        no_move: bool
            If True, the project in the configuration files is not moved to the
            position right before the configuration of the current project
        """
        config = self.config
        if projectname is None:
            projectname = config.setdefault(
                'project', 'sensitivity_' + str(self.organizer.experiment))
        self.projectname = projectname
        if root_dir is None:
            root_dir = self.project_config.get(
                'root', osp.dirname(self.organizer.project_config['root']))
        self.project_config['root'] = root_dir
        organizer = self.sa_organizer
        organizer.start(setup=dict(
            root_dir=root_dir, projectname=projectname, link=link,
            src_project=self.organizer.projectname))
        if not six.PY2 and not no_move:
            utils.ordered_move(organizer.config.projects, projectname,
                               self.organizer.projectname)

    def compile_model(self):
        """Compile the sensitivity analysis model"""
        self.sa_organizer.start(compile_model=dict(
            projectname=self.projectname))

    def init(self, nml=None, experiment=None, run_prepare=False,
             use_param=False, no_move=False):
        """
        Initialize the experiments for the sensitivity analysis

        Parameters
        ----------
        nml: dict
            A mapping from namelist parameters to the values to use. Ranges
            might be lists of numbers or ``'<i>err'`` to use ``'<i>'``-times
            the error from the parameterization. You might also provide up to
            three values in case on of them is a string with ``'err'`` in it,
            where the first value corresponds to the minimum, the second to
            the maximum and the third to the number of steps.
        experiment: str
            The base experiment name to use. Should somehow contain a number
            which will be increased for each experiment. If None, it will
            default to ``'<id>_sens0'``
        run_prepare: bool
            If True or no ``'input'`` key exists for the experiment, the
            ``'prepare'`` evaluation task is run to get the reference data.
            Otherwise, if False and the ``'input'`` key does not exist, the
            specified file is used
        use_param: bool
            If True/set and the parameterization is required, use the
            parameterization of the original experiment
        no_move: bool
            If True, the experiment in the configurations are not moved to the
            position right before the configuration of the current experiment
        """
        def get_err_range(error, nml, key, values):
            """Compute the ranges for the given namelist `key` based upon the
            mean"""
            converted = [0] * len(values)
            for i, val in enumerate(values):
                if isstring(val) and 'err' in val:
                    m = utils.float_patt.search(val)
                    multiplier = float(m.group()) if m else 1.0
                    converted[i] = nml[key] + multiplier * errors[key]
                else:
                    converted[i] = float(val)
            return list(map(float, np.linspace(*converted, endpoint=True)))

        def setup_paramdir(experiment):
            """Link files from the base experiment param dir to the new one"""
            paramdir = organizer.exp_config.setdefault(
                    'paramdir', osp.join(self.organizer.exp_config['expdir'],
                                         'parameterization'))
            root_dir = self.exp_config.get('paramdir')
            if root_dir and osp.exists(root_dir):
                for f in os.listdir(root_dir):
                    organizer._link(osp.join(root_dir, f),
                                    osp.join(paramdir, f))
            if self.exp_config.get('database'):
                organizer.exp_config['database'] = self.exp_config[
                    'database']

        def update_nml():
            for nml_key, settings in self.exp_config.get(
                    'namelist', {}).items():
                for key, val in settings.items():
                    organizer.exp_config['namelist'][nml_key].setdefault(
                        key, val)

        def transposed_dict(iterable):
            return map(OrderedDict, starmap(zip, zip(
                       repeat(errs.keys()), product(*iterable))))

        logger = self.logger

        if nml is None:
            nml = self.config.get('namelist', self.project_config.get(
                'sensitivity_namelist'))
        if nml is None:
            raise ValueError(
                "No parameter ranges specified! Please use the params "
                "parameter!")
        nml = OrderedDict(nml)
        errs = OrderedDict()
        ranges = OrderedDict()
        for key, val in nml.items():
            if any(isstring(s) and 'err' in s for s in val):
                errs[key] = val
            else:
                ranges[key] = list(map(float, val))
        nml = OrderedDict(chain(ranges.items(), errs.items()))
        self.config['namelist'] = nml.copy()
        self.project_config['sensitivity_namelist'] = nml.copy()
        try:
            eval_stations = self.exp_config['eval_stations']
        except KeyError:
            raise ValueError("No evaluation stations specified for the "
                             "experiment %s!" % (self.experiment))
        if errs:
            from gwgen.parameterization import Parameterizer
            param_stations = self.exp_config.get('param_stations')
            if param_stations is None:
                raise ValueError("No parameterization stations specified for "
                                 "the experiment %s!" % (self.experiment))
            tasks = list(map(Parameterizer.get_task_for_nml_key, errs))
            other_tasks = list(map(Parameterizer.get_task_for_nml_key, ranges))
            err_bins = OrderedDict()
            flags = OrderedDict()
            flags_no_mean = OrderedDict()
        else:
            tasks = []
        organizer = self.sa_organizer
        if experiment is None:
            experiment = self.experiment + '_sens0'

        #: base string for the experiment description. The values are inserted
        #: later
        base_description = ('Sensitivity analysis of %s with %s') % (
            self.experiment, ', '.join('%s={}' % key for key in chain(
                ranges, errs)))
        # choose if we have to evaluate the data
        run_prepare = run_prepare or {'input', 'reference'}.difference(
            self.exp_config)
        if not run_prepare:
            input_file = self.exp_config['input']
            reference = self.exp_config['reference']

        for i, d in enumerate(map(lambda t: OrderedDict(zip(*t)), zip(
                repeat(ranges.keys()), product(*ranges.values())))):
            experiment = utils.get_next_name(experiment)
            organizer.init(projectname=self.projectname, experiment=experiment)
            if not six.PY2 and not no_move:
                utils.ordered_move(organizer.config.experiments, experiment,
                                   self.experiment)
            organizer.exp_config['namelist'] = {'weathergen_ctl': d.copy()}
            if run_prepare:
                organizer.start(evaluate=dict(
                    stations=eval_stations, prepare={'to_csv': True},
                    experiment=experiment))
                input_file = organizer.exp_config['input']
                reference = organizer.exp_config['reference']
                run_prepare = False
            organizer.exp_config['input'] = input_file
            organizer.exp_config['reference'] = reference
            organizer.exp_config['eval_stations'] = eval_stations
            organizer.exp_config['base_exp'] = last = organizer.experiment
            if use_param:
                setup_paramdir(experiment)
            update_nml()
            if not tasks:
                organizer.exp_config['description'] = base_description.format(
                    *d.values())
            else:  # run the parameterization
                organizer.exp_config['param_stations'] = param_stations
                param_config = {t.name: {} for t in tasks}
                for t, (nml_key, val) in zip(other_tasks, d.items()):
                    config_key = t.get_config_key(nml_key)
                    if config_key:
                        param_config[t.name] = {config_key: val}
                organizer.fix_paths(organizer.exp_config)
                organizer.fix_paths(organizer.project_config)

                logger.info('Parameterizing experiment %i', i)
                manager = organizer.start(param=dict(
                    experiment=experiment, **param_config)).param
                param_tasks = {t.name: t for t in manager.tasks}
                nml = organizer.exp_config['namelist']['weathergen_ctl']

                # insert description
                organizer.exp_config['description'] = base_description.format(
                    *chain(d.values(), map(nml.get, errs)))

                # insert meta information that makes it easier to categorize
                # the experiments later
                organizer.exp_config['flags'] = OrderedDict([
                    (key, 'mean') for key in errs])

                errors = {key: param_tasks[t.name].get_error(key)
                          for key, t in zip(errs, tasks)}
                func = partial(get_err_range, errors, nml)
                # Convert the error ranges into lists
                err_ranges = list(starmap(func, errs.items()))
                logger.debug('Error ranges:')
                for key, vals in zip(errs, err_ranges):
                    if not len(vals) % 2:  # insert mean for configuration
                        vals.insert(len(vals) // 2, nml[key])
                    if i == 0:
                        err_bins[key] = np.zeros(
                            (np.product(list(map(len, ranges.values()))),
                             len(vals)))
                        flags[key] = list(map(
                            lambda f: '%4.2ferr' % f, get_err_range(
                                {key: 1}, {key: 0}, key, errs[key])))
                        flags_no_mean[key] = flags[key][:]
                        # insert (or rename) mean for configuration
                        if not len(flags[key]) % 2:
                            flags[key].insert(len(flags[key]) // 2, 'mean')
                        else:
                            flags[key][(len(flags[key]) - 1) // 2] = 'mean'
                            flags_no_mean[key].pop(len(flags[key]) // 2)
                    err_bins[key][i] = vals
                    logger.debug('    %s: %s', key, vals)
                    vals.pop((len(vals) - 1) // 2)  # delete the mean again

                # split them up and use them as namelist input
                # note: d2 is the extended namelist for each experiment
                for d2, fl in zip(transposed_dict(err_ranges), transposed_dict(
                        flags_no_mean.values())):
                    experiment = utils.get_next_name(experiment)
                    combined = OrderedDict(chain(d.items(), d2.items()))
                    organizer.start(init=dict(
                        projectname=self.projectname, experiment=experiment,
                        description=base_description.format(*combined)))
                    if not six.PY2 and not no_move:
                        utils.ordered_move(
                            organizer.config.experiments, experiment,
                            self.experiment)
                    organizer.exp_config['namelist'] = {
                        'weathergen_ctl': combined}
                    update_nml()
                    organizer.exp_config['input'] = input_file
                    organizer.exp_config['reference'] = reference
                    organizer.exp_config['eval_stations'] = eval_stations
                    organizer.exp_config['base_exp'] = last
                    organizer.exp_config['flags'] = fl
        if errs:
            self.config['unstructured'] = uconf = OrderedDict()
            for err, arr in err_bins.items():
                minmeanmax = np.zeros((arr.shape[1], 3))
                minmeanmax[:, 0] = arr.min(axis=0)
                minmeanmax[:, 1] = arr.mean(axis=0)
                minmeanmax[:, 2] = arr.max(axis=0)
                uconf[err] = OrderedDict(zip(flags[key], minmeanmax.tolist()))
                logger.debug('%s', err)
                logger.debug('   minima: %s', minmeanmax[:, 0])
                logger.debug('     mean: %s', minmeanmax[:, 1])
                logger.debug('   maxima: %s', minmeanmax[:, 2])

    docstrings.keep_params('GWGENOrganizer.run.parameters', 'remove')
    docstrings.keep_params(
        'SensitivityAnalysis._parallelize_command.parameters', 'experiments')

    @docstrings.dedent
    def run(self, experiments=None, remove=False):
        """
        Run the analysis

        Parameters
        ----------
        %(SensitivityAnalysis._parallelize_command.parameters.experiments)s
        %(GWGENOrganizer.run.parameters.remove)s"""
        self._parallelize_command(dict(run=dict(remove=remove)),
                                  experiments=experiments)

    docstrings.keep_params(
        'SensitivityAnalysis._parallelize_command.parameters', 'loop_exps')

    @docstrings.dedent
    def evaluate(self, experiments=None, loop_exps=False, **kwargs):
        """
        Evaluate the experiments of the sensitivity analysis

        Parameters
        ----------
        %(SensitivityAnalysis._parallelize_command.parameters.experiments)s
        %(SensitivityAnalysis._parallelize_command.parameters.loop_exps)s
        %(GWGENOrganizer.evaluate.parameters)s
        """
        for key, val in kwargs.items():
            if isinstance(val, Namespace):
                kwargs[key] = val = vars(val)
            if isinstance(val, dict):
                val.pop('experiments', None)
                val.pop('loop_exps', None)
        self._parallelize_command(dict(evaluate=kwargs),
                                  experiments=experiments,
                                  loop_exps=loop_exps)

    @docstrings.dedent
    def plot(self, indicators=['rsquared', 'slope', 'ks', 'quality'],
             names=['prcp', 'tmin', 'tmax', 'mean_cloud', 'wind'],
             meta=None, **kwargs):
        """
        Plot the result of the sensitivity analysis

        2 different plots are made for each variable and quality indicator:

            1. 1d plots with the indicator on the y-axis and the namelist
               parameter on the x-axis
            2. 2d plots with one indicator on the x-axis and another on the
               y-axis for each possible combination

        You can disable one of the plot types via the `only_1d` and `only_2d`
        parameter.

        Parameters
        ----------
        indicators: str or list of str
            The name of the indicators from the `quality` evaluation task
        names: str or list of str
            The name(s) of the variable(s) to plot
        meta: dict or path to yaml configuration file
            Alternative meta information for the data set
        ``**kwargs``
            Any task of the :class:`SensitivityPlot` framework
        """
        meta = meta or self.config.get('additional_meta')
        if isinstance(meta, six.string_types):
            with open(meta) as f:
                meta = yaml.load(f)
        if meta is not None:
            self.config['additional_meta'] = meta
        for task, d in list(kwargs.items()):
            d = kwargs[task] = vars(d) if isinstance(d, Namespace) else dict(d)
            d['names'] = names
            d['indicators'] = indicators
            d['meta'] = meta
            d['config'] = self.exp_config
            d['project_config'] = self.organizer.project_config
            d['sa'] = self
        manager = SensitivityPlot.get_manager(
            config=self.organizer.global_config)
        manager.initialize_tasks(np.array(self.experiments), kwargs)
        manager.setup(np.array(self.experiments))
        manager.run(self.config)
        return manager

    def remove(self, *args, **kwargs):
        """
        Remove the project and files of the sensitivity analysis"""
        self.organizer.start(
            remove=dict(projectname=self.projectname, complete=True,
                        *args, **kwargs))


_SensitivityPlotConfig = namedtuple(
    '_SensitivityPlotConfig', ('sa', 'indicators', 'names', 'meta'))

_SensitivityPlotConfig = utils.append_doc(
    _SensitivityPlotConfig, docstrings.get_sections("""
Parameters
----------
sa: SensitivityAnalysis
    The analyser to use
indicators: str or list of str
    The name of the indicators from the `quality` evaluation task
names: str or list of str
    The name of the variables to plot
meta: dict
    Alternative meta information for the data set
""", '_SensitivityPlotConfig'))

SensitivityPlotConfig = utils.enhanced_config(_SensitivityPlotConfig,
                                              'SensitivityPlotConfig')


@docstrings.dedent
def default_sens_config(
        sa=None, indicators=['rsquared', 'slope', 'ks', 'quality'],
        names=['prcp', 'tmin', 'tmax', 'mean_cloud', 'wind'],
        meta={}, *args, **kwargs):
    """
    The default configuration for :class:`SensitivityPlot` instances.
    See also the :attr:`SensitivityPlot.default_config` attribute

    Parameters
    ----------
    %(SensitivityPlotConfig.parameters)s"""
    return SensitivityPlotConfig(
            sa, indicators, names, meta,
            *utils.default_config(*args, **kwargs))


class SensitivityPlot(utils.TaskBase):
    """A base class for the plots of the :class:`SensitivityAnalysis`"""

    _registry = []

    meta = {
        'rsquared': dict(long_name='Mean R$^2$ of all quantiles'),
        'slope': dict(long_name='Slope indicator'),
        'quality': dict(long_name='Quality indicator'),
        'ks': dict(
            long_name='Fraction of stations without significant difference'),
        'thresh': dict(long_name='Crossover point of Gamma-GP distribution',
                       units='mm'),
        'gp_shape': dict(long_name='Generalized Pareto Shape parameter',
                         units='-')
        }

    variables_meta = {
        'prcp': 'Precipitation',
        'tmin': 'Min. Temperature',
        'tmax': 'Max. Temperature',
        'mean_cloud': 'Cloud fraction'}

    _datafile = 'sensitivity_analysis.csv'

    dbname = 'sensitivity_analysis'

    @property
    def task_data_dir(self):
        """The directory where to store data"""
        return self.sa_dir

    @property
    def namelist(self):
        """The namelist of the sensitivity analysis holding the settings"""
        return self.task_config.sa.config['namelist']

    @property
    def err_nml_keys(self):
        return [key for key, l in self.namelist.items() if any(
            isstring(s) and 'err' in s for s in l)]

    @property
    def ranges_nml_keys(self):
        errs = self.err_nml_keys
        return [key for key in self.namelist if key not in errs]

    @property
    def organizer(self):
        return self.task_config.sa.organizer

    @property
    def base_experiments(self):
        all_exps = self.organizer.config.experiments
        return [exp for exp in self.stations
                if all_exps[exp].get('base_exp') == exp]

    @property
    def default_config(self):
        return default_sens_config()._replace(
            **super(SensitivityPlot, self).default_config._asdict())

    @classmethod
    def _modify_parser(cls, parser):
        parser.setup_args(default_sens_config)
        parser, setup_grp, run_grp = super(
            SensitivityPlot, cls)._modify_parser(parser)
        parser.pop_arg('names')
        parser.pop_arg('indicators')
        parser.pop_arg('sa')
        parser.pop_arg('meta')
        return parser, setup_grp, run_grp

    @property
    def sql_dtypes(self):
        import sqlalchemy
        ret = super(SensitivityPlot, self).sql_dtypes
        for indicator in ['rsquared', 'slope', 'ks', 'quality'] + list(
                self.namelist.keys()):
            ret[indicator] = sqlalchemy.REAL
        ret['vname'] = sqlalchemy.TEXT
        return ret

    def setup_from_file(self, *args, **kwargs):
        kwargs['index_col'] = ['id', 'vname'] + list(self.namelist.keys())
        return super(SensitivityPlot, self).setup_from_file(*args, **kwargs)

    def setup_from_db(self, *args, **kwargs):
        kwargs['index_col'] = ['id', 'vname'] + list(self.namelist.keys())
        return super(SensitivityPlot, self).setup_from_db(*args, **kwargs)

    def add_meta(self, ds):
        meta = self.meta.copy()
        meta.update(self.task_config.meta)
        for key, attrs in meta.items():
            if key in ds:
                ds[key].attrs.update(attrs)
            if key + '_flag' in ds:
                ds[key + '_flag'].attrs.update(attrs)

    @property
    def ds(self):
        """Gives the base dataset for the plot"""
        import xarray as xr
        ds = xr.Dataset.from_dataframe(self.data.reset_index('id'))
        self.add_meta(ds)
        return ds

    def setup_from_scratch(self):
        import pandas as pd
        experiments = self.stations
        all_exps = self.organizer.config.experiments
        dfs = []
        for variable in self.task_config.names:
            data = [{} for _ in experiments]
            for i, exp in enumerate(experiments):
                try:
                    data[i] = all_exps[exp]['evaluation']['quality'][variable]
                except KeyError:
                    pass
            df = pd.DataFrame(
                data, index=pd.Index(experiments, name='id'))
            df2 = pd.DataFrame(
                [all_exps[exp]['namelist']['weathergen_ctl']
                 for exp in experiments],
                index=pd.Index(experiments, name='id'))
            merged = df.merge(df2, left_index=True, right_index=True)
            if self.err_nml_keys:
                flags_df = pd.DataFrame(
                    [all_exps[exp]['flags'] for exp in experiments],
                    index=pd.Index(experiments, name='id'))
                merged = merged.merge(
                    flags_df, left_index=True, right_index=True,
                    suffixes=('', '_flag'))
            merged['vname'] = variable
            errs_keys = [key + '_flag' for key in self.err_nml_keys]
            dfs.append(merged.set_index(
                ['vname'] + self.ranges_nml_keys + errs_keys,
                append=True))
        self.data = pd.concat(dfs)


class SensitivityPlot1D(SensitivityPlot):
    """1d plots with the indicator on the y-axis and the namelist parameter on
    the x-axis"""

    name = 'plot1d'

    summary = ('1d plots with the indicator on the y-axis and the namelist '
               'parameter on the x-axis')

    has_run = True

    fmt = dict(
        xlabel='{desc}',
        ylabel='%(long_name)s',
        title='%(vname_long)s',
        plot=' ',
        marker='o',
        color='coolwarm')

    def create_project(self, ds):
        import psyplot.project as psy

        def get_key(nml_key):
            return nml_key if nml_key not in errs else nml_key + '_flag'
        vmeta = self.variables_meta
        variables = ds.vname.values
        indicators = self.task_config.indicators
        nml = self.namelist
        errs = self.err_nml_keys
        vmeta = OrderedDict([(key, vmeta[key]) for key in ds.vname.values])
        for nml_key in nml:
            for i, variable in enumerate(variables):
                for ind in indicators:
                    other_keys = {key: range(ds[key].size)
                                  for key in map(get_key,
                                                 set(nml) - {nml_key})}
                    label = ', '.join(
                        map('{0}=%({0})s'.format, other_keys))
                    other_keys['vname'] = i
                    fmt = self.fmt.copy()
                    psy.plot.lineplot(
                        ds, name=ind, dims=other_keys, fmt=fmt,
                        coord=nml_key, legendlabels=label,
                        attrs={'vname_long': vmeta.get(variable)})
        return psy.gcp(True)[:]


_SensitivityPlot2DConfig = namedtuple(
    '_SensitivityPlot2DConfig',
    ['logx', 'logy'] + list(_SensitivityPlotConfig._fields))

_SensitivityPlot2DConfig = utils.append_doc(
    _SensitivityPlot2DConfig, docstrings.get_sections(docstrings.dedents("""
Parameters
----------
logx: bool
    If True, the x-axis will be logarithmic
logy: bool
    If True, the y-axis will be logarithmic
%(_SensitivityPlotConfig.parameters)s
"""), '_SensitivityPlot2DConfig'))

SensitivityPlot2DConfig = utils.enhanced_config(_SensitivityPlot2DConfig,
                                                'SensitivityPlot2DConfig')


@docstrings.dedent
def default_plot2d_config(
        logx=False, logy=False, *args, **kwargs):
    """
    The default configuration for :class:`SensitivityPlot` instances.
    See also the :attr:`SensitivityPlot.default_config` attribute

    Parameters
    ----------
    %(SensitivityPlot2DConfig.parameters)s"""
    return SensitivityPlot2DConfig(
            logx, logy, *default_sens_config(*args, **kwargs))


class SensitivityPlot2D(SensitivityPlot):
    """2d plots with one namelist parameter on the y-axis and one on the
    x-axis and a color coding the the indicator"""

    name = 'plot2d'

    summary = ('2d plots with one namelist parameter on the y-axis and one on '
               'the x-axis and a color coding the the indicator')

    has_run = True

    @property
    def default_config(self):
        return default_plot2d_config()._replace(
            **super(SensitivityPlot2D, self).default_config._asdict())

    @classmethod
    def _modify_parser(cls, parser):
        parser.setup_args(default_plot2d_config)
        parser, setup_grp, run_grp = super(
            SensitivityPlot2D, cls)._modify_parser(parser)
        parser.update_arg('logx', group=run_grp)
        parser.update_arg('logy', group=run_grp)
        return parser, setup_grp, run_grp

    def run(self, info):
        import psyplot.data as psyd
        import matplotlib.pyplot as plt
        import psy_simple.plotters as psyps
        import matplotlib.colors as mcol
        import seaborn as sns
        from matplotlib.backends.backend_pdf import PdfPages
        import matplotlib.patches as patches
        sns.set_style('white')
        full_df = self.data.reset_index()
        plot_output = self.task_config.plot_output
        pdf = PdfPages(plot_output)
        all_exps = self.organizer.config.experiments
        for ind in self.task_config.indicators:
            for variable in self.task_config.names:
                df = full_df[full_df.vname == variable]
                vals = {
                    key: group.set_index('gp_shape').sort_index()[ind]
                    for key, group in df.groupby('thresh')}
                y = psyd._infer_interval_breaks(sorted(vals.keys()))
                fig = plt.figure(edgecolor='none', facecolor='none')
                if mpl.__version__ < '2.0':
                    ax = plt.axes(axisbg='0.9')
                else:
                    ax = plt.axes(facecolor='0.9')
                all_vals = np.concatenate(list(vals.values()))
                vmin, vmax = psyps.DataTicksCalculator._round_min_max(
                    np.nanmin(all_vals), np.nanmax(all_vals))
                norm = mcol.BoundaryNorm(
                    np.linspace(vmin, vmax, 11, endpoint=True), 10)
                cmap = plt.get_cmap('Greens', 10)
                # set missing values to (dark) red
                cmap.set_bad((0.7686274509803922, 0.3058823529411765,
                              0.3215686274509804))
                plots = []
                x_vals = {}
                for i, key in enumerate(sorted(df.thresh.unique())):
                    data = np.ma.array(vals[key], mask=np.isnan(vals[key]))
                    x_vals[key] = x = psyd._infer_interval_breaks(
                        vals[key].index.values)
                    plots.append(ax.pcolormesh(
                        x, y[i:i+2], data[np.newaxis, :], norm=norm,
                        cmap=cmap))
                ax.set_xlabel('Generalized Pareto shape parameter')
                ax.set_ylabel('Gamma-GP crossover point [mm]')
                ax.set_title(self.variables_meta[variable])
                cbar = plt.colorbar(plots[-1], orientation='horizontal')
                cbar.set_label(self.meta[ind]['long_name'])
                if self.task_config.logx:
                    ax.set_xscale('log')
                if self.task_config.logy:
                    ax.set_yscale('log')
                # draw box around colorbar
                plt.rcParams['axes.edgecolor'] = 'k'
                if mpl.__version__ < '2.0':
                    ax = plt.axes([0.08, 0.03, 0.85, 0.22], axisbg='none')
                else:
                    ax = plt.axes([0.08, 0.03, 0.85, 0.22], fc='none')
                ax.grid(False)
                ax.set_xticklabels([])
                ax.set_yticklabels([])
                ax.add_patch(patches.Rectangle((0, 0), 1, 1, fill=False,
                                               edgecolor='k', linewidth=2))
                if not self.err_nml_keys:
                    pdf.savefig(fig)
                    continue
                mean_boxes = []
                cmap = plt.get_cmap('coolwarm', 2)
                means = {}
                medians = {}  # XXX currently not used
                for exp in self.base_experiments:
                    nml = all_exps[exp]['namelist']['weathergen_ctl']
                    param_config = all_exps[exp]['parameterization']['prcp']
                    means[nml['thresh']] = param_config['gpshape_mean']
                    medians[nml['thresh']] = param_config['gpshape_median']
                for i, key in enumerate(sorted(df.thresh.unique())):
                    mean_index = df.gp_shape[
                        df.thresh == key].values.searchsorted(means[key])
                    mean_boxes.append(patches.Rectangle(
                        [x_vals[key][mean_index], y[i]],
                        x_vals[key][mean_index + 1] - x_vals[key][mean_index],
                        y[i + 1] - y[i], edgecolor=cmap(0), facecolor='none',
                        lw=2.0))
                    ax.add_patch(mean_boxes[-1])
                plt.legend([mean_boxes[0]],
                           ['mean'], title='GP shape parameter')
                pdf.savefig(fig)
        pdf.close()
        info['plot_file'] = plot_output



# -*- coding: utf-8 -*-
"""Additional routines for preprocessing"""
import tempfile
import os.path as osp
from collections import namedtuple
import numpy as np
import pandas as pd
from psyplot.compat.pycompat import OrderedDict
import gwgen.utils as utils
from gwgen.utils import docstrings


class CloudPreproc(utils.TaskBase):

    @property
    def task_data_dir(self):
        return osp.join(self.data_dir, 'eecra')

    _registry = []


class CloudInventory(CloudPreproc):
    """A task for computing the EECRA inventory for each station"""

    name = 'eecra_inventory'

    summary = 'Compute the inventory of the EECRA stations'

    http_xstall = 'http://cdiac.ornl.gov/ftp/ndp026c/XSTALL'

    _datafile = 'eecra_inventory.csv'

    dbname = 'eecra_inventory'

    has_run = True

    @property
    def setup_parallel(self):
        return self.setup_from == 'scratch'

    @property
    def default_config(self):
        return default_cloud_inventory_config()._replace(
            **super(CloudInventory, self).default_config._asdict())

    @property
    def xstall_df(self):
        """The dataframe corresponding to the XSTALL stations"""
        use_xstall = self.task_config.xstall
        if utils.isstring(use_xstall):
            fname = self.task_config.no_xstall
        else:
            fname = tempfile.NamedTemporaryFile().name
            utils.download_file(self.http_xstall, fname)
        arr = np.loadtxt(fname, usecols=[1, 2, 3])
        df = pd.DataFrame(arr, columns=['station_id', 'lat', 'lon'])
        df['station_id'] = df.station_id.astype(int)
        df.set_index('station_id', inplace=True)
        return df

    @classmethod
    def _modify_parser(cls, parser):
        parser.setup_args(default_cloud_inventory_config)
        cls.has_run = False
        parser, setup_grp, run_grp = super(CloudInventory, cls)._modify_parser(
            parser)
        parser.update_arg('xstall', group=setup_grp)
        cls.has_run = True
        return parser, setup_grp, run_grp

    def __init__(self, *args, **kwargs):
        super(CloudInventory, self).__init__(*args, **kwargs)
        self.__setup = False

    def setup(self, *args, **kwargs):
        self.__setup = True
        super(CloudInventory, self).setup(*args, **kwargs)

    def init_from_scratch(self):
        from gwgen.parameterization import HourlyCloud
        task = HourlyCloud.from_task(self)
        task.download_src(task.raw_dir)  # make sure the source files exist

    def setup_from_file(self, **kwargs):
        """Set up the task from already stored files (and avoid locating the
        stations of this task)"""
        kwargs = self._split_kwargs(kwargs)
        for i, datafile in enumerate(utils.safe_list(self.datafile)):
            self._set_data(pd.read_csv(datafile, **kwargs[i]), i)

    def setup_from_db(self, **kwargs):
        """Set up the task from datatables already created (and avoid locating
        the stations of this task)"""
        kwargs = self._split_kwargs(kwargs)
        for i, dbname in enumerate(utils.safe_list(self.dbname)):
            self._set_data(pd.read_sql_query(
                "SELECT * FROM %s" % (self.dbname),
                self.engine, **kwargs[i]), i)

    def setup_from_scratch(self):
        from gwgen.parse_eecra import parse_file

        def compute(fname):
            g = parse_file(fname).groupby('station_id')[
                ['lat', 'lon', 'year']]
            df = g.mean()
            df['counts'] = g.size()
            std = g.std()
            df['lon_std'] = std.lon
            df['lat_std'] = std.lat
            return df

        self.data = pd.concat(list(map(compute, self.stations)))

    def write2db(self, *args, **kwargs):
        if self.__setup:
            return
        kwargs.setdefault('if_exists', 'replace')
        super(CloudInventory, self).write2db(*args, **kwargs)

    def write2file(self, *args, **kwargs):
        if self.__setup:
            return
        super(CloudInventory, self).write2file(*args, **kwargs)

    def run(self, info):

        self.__setup = False

        if self.setup_from == 'scratch':
            df = self.data
            # we may use a parallel setup which requires a weighted average
            g = df.groupby(level='station_id')
            total_counts = g.counts.transform("sum")
            df['lat'] = df.counts / total_counts * df.lat
            df['lon'] = df.counts / total_counts * df.lon
            df['lat_std'] = (df.counts / total_counts) * df.lat_std ** 2
            df['lon_std'] = (df.counts / total_counts) * df.lon_std ** 2
            eecra = g.agg(OrderedDict([
                    ('lat', 'sum'), ('lon', 'sum'), ('lat_std', 'sum'),
                    ('lon_std', 'sum'),
                    ('year', ('min', 'max')), ('counts', 'sum')]))
            eecra.columns = ['lat', 'lon', 'lat_std', 'lon_std',
                             'firstyear', 'lastyear', 'counts']
            eecra[['lat_std', 'lon_std']] **= 0.5

            use_xstall = self.task_config.xstall

            if use_xstall:
                to_replace = self.xstall_df
                # keep only matching stations
                to_replace = to_replace.join(eecra[[]], how='inner')
                eecra.loc[to_replace.index, ['lat', 'lon']] = to_replace
            self.data = eecra

        if self.task_config.to_csv:
            self.write2file()
        if self.task_config.to_db:
            self.write2db()


class CloudGHCNMap(CloudPreproc):
    """A task for computing the EECRA inventory for each station"""

    name = 'eecra_ghcn_map'

    setup_requires = ['eecra_inventory']

    summary = 'Compute the inventory of the EECRA stations'

    _datafile = 'eecra_ghcn_map.csv'

    dbname = 'eecra_ghcn_map'

    has_run = True

    @property
    def default_config(self):
        return default_cloud_ghcn_map_config()._replace(
            **super(CloudGHCNMap, self).default_config._asdict())

    @classmethod
    def _modify_parser(cls, parser):
        parser.setup_args(default_cloud_ghcn_map_config)
        cls.has_run = False
        parser, setup_grp, run_grp = super(CloudGHCNMap, cls)._modify_parser(
            parser)
        parser.update_arg('max_distance', group=setup_grp, short='md')
        parser.pop_arg('to_db')
        parser.pop_arg('setup_from')
        cls.has_run = True
        return parser, setup_grp, run_grp

    def __init__(self, *args, **kwargs):
        super(CloudGHCNMap, self).__init__(*args, **kwargs)
        self.task_config = self.task_config._replace(
            setup_from='scratch', to_db=False)
        self.__setup = False

    def setup(self, *args, **kwargs):
        self.__setup = True
        super(CloudGHCNMap, self).setup(*args, **kwargs)

    def init_from_scratch(self):
        from gwgen.parameterization import HourlyCloud
        task = HourlyCloud.from_task(self)
        task.download_src(task.raw_dir)  # make sure the source files exist

    def setup_from_scratch(self):
        """Does nothing but initializing an empty data frame. The real work is
        done in the :meth:`run` method"""
        self.data = pd.DataFrame([], columns=['station_id', 'dist'],
                                 index=pd.Index([], name='id'))

    def write2db(self, *args, **kwargs):
        raise NotImplementedError(
            'The data is always written to the database!')

    def write2file(self, *args, **kwargs):
        if self.__setup:
            return
        super(CloudGHCNMap, self).write2file(*args, **kwargs)

    def run(self, info):
        def create_geog(table):
            conn.execute(
                'ALTER TABLE %s ADD COLUMN geog geography(POINT,4326) ;' % (
                    table))
            conn.execute("""
                UPDATE %s
                    SET geog = ST_SetSRID(ST_MakePoint(lon,lat),4326);""" % (
                    table))
            conn.execute(
                'CREATE INDEX {0}_geog ON {0} USING GIST (geog);'.format(table)
                )
        from gwgen.evaluation import EvaluationPreparation
        self.__setup = False

        inv = self.eecra_inventory

        if not self.engine.has_table(inv.dbname):
            inv.write2db()
        conn = self.engine.connect()
        if 'geog' not in pd.read_sql_query('SELECT * FROM %s LIMIT 0' % (
                inv.name), conn).columns:
            create_geog(inv.dbname)

        t = EvaluationPreparation.from_task(self)
        # download inventory
        t.download_src()
        ghcn = t.station_list
        ghcn = ghcn.ix[ghcn.vname == 'PRCP'].set_index('id')
        ghcn.to_sql('ghcn_inventory', self.engine, if_exists='replace')
        create_geog('ghcn_inventory')

        conn.execute("DROP TABLE IF EXISTS eecra_ghcn_map;")

        conn.execute("""
            CREATE TABLE eecra_ghcn_map AS (
                SELECT DISTINCT ON (id) id, station_id, dist FROM (
                    SELECT DISTINCT ON (a.station_id)
                        b.id, a.station_id, ST_Distance(a.geog, b.geog) AS dist
                    FROM eecra_inventory a
                        INNER JOIN ghcn_inventory b ON ST_DWithin(
                            a.geog, b.geog, 1000)
                    ORDER BY a.station_id, ST_Distance(a.geog, b.geog)) foo
                ORDER BY id, dist);""")

        self.data = pd.read_sql('eecra_ghcn_map', self.engine, index_col='id')
        conn.close()

        if self.task_config.to_csv:
            self.write2file()


CloudInventoryConfig = namedtuple(
    'CloudInventoryConfig', ['xstall'] + list(utils.TaskConfig._fields))


# to_db is set to True by default because it is required
docstrings.delete_params('TaskConfig.parameters', 'to_db', 'setup_from')


CloudInventoryConfig = utils.append_doc(
    CloudInventoryConfig, docstrings.get_sections(docstrings.dedents("""
    Parameters
    ----------
    xstall: bool or str
        If True (default), download the XSTALL file from %s.
        This file contains some estimates of station longitude and latitude.
        If ``False`` or empty string, the file is not used, otherwise, if set
        with a string, it is interpreted as the path to the local file
    %%(TaskConfig.parameters.no_to_db|setup_from)s
    """ % CloudInventory.http_xstall), 'CloudInventoryConfig'))


@docstrings.dedent
def default_cloud_inventory_config(xstall=True, *args, **kwargs):
    """
    Default config for :class:`CloudInventory`

    Parameters
    ----------
    %(CloudInventoryConfig.parameters)s"""
    return CloudInventoryConfig(
        xstall, *utils.default_config(*args, **kwargs))


CloudGHCNMapConfig = namedtuple(
    'CloudGHCNMapConfig', ['max_distance'] + list(utils.TaskConfig._fields))


# to_db is set to True by default because it is required
docstrings.delete_params('TaskConfig.parameters', 'to_db', 'setup_from')


CloudGHCNMapConfig = utils.append_doc(
    CloudGHCNMapConfig, docstrings.get_sections(docstrings.dedents("""
    Parameters
    ----------
    max_distance: float
        The maximum distance in meters for which we consider two stations as
        equal (Default: 1000m)
    %(TaskConfig.parameters.no_to_db|setup_from)s
    """), 'CloudGHCNMapConfig'))


@docstrings.dedent
def default_cloud_ghcn_map_config(max_distance=1000., *args, **kwargs):
    """
    Default config for :class:`CloudGHCNMap`

    Parameters
    ----------
    %(CloudGHCNMapConfig.parameters)s"""
    return CloudGHCNMapConfig(
        max_distance, *utils.default_config(*args, **kwargs))



from __future__ import print_function, division
import os
import os.path as osp
import six
import re
import shutil
import sys
import datetime as dt
from itertools import repeat
from argparse import Namespace, RawTextHelpFormatter
import logging
import numpy as np
import gwgen.utils as utils
from gwgen.utils import docstrings
from model_organization import ModelOrganizer
from model_organization.config import ordered_yaml_dump

from collections import OrderedDict


class GWGENOrganizer(ModelOrganizer):
    """
    A class for organizing a model

    This class is indended to have hold the basic functions for organizing a
    model. You can subclass the functions ``setup, init`` to fit to your model.
    When using the model from the command line, you can also use the
    :meth:`setup_parser` method to create the argument parsers"""

    commands = ModelOrganizer.commands
    commands.insert(commands.index('init'), 'compile_model')
    commands.insert(commands.index('archive'), 'preproc')
    commands.insert(commands.index('archive'), 'param')
    commands.insert(commands.index('archive'), 'run')
    commands.insert(commands.index('archive'), 'evaluate')
    commands.insert(commands.index('archive'), 'bias_correction')
    commands.insert(commands.index('archive'), 'sensitivity_analysis')

    #: mapping from the name of the parser command to the method name
    parser_commands = {'compile_model': 'compile',
                       'sensitivity_analysis': 'sens',
                       'bias_correction': 'bias'}

    #: list of str. The keys describing paths for the model
    paths = ['expdir', 'src', 'data', 'param_stations', 'eval_stations',
             'indir', 'input', 'outdir', 'outdata', 'nc_file',  'project_file',
             'plot_file', 'reference', 'evaldir', 'paramdir', 'workdir',
             'param_grid', 'grid', 'eval_grid']

    name = 'gwgen'

    # -------------------------------------------------------------------------
    # --------------------------- Infrastructure ------------------------------
    # ---------- General parts for organizing the model infrastructure --------
    # -------------------------------------------------------------------------

    docstrings.get_sectionsf('ModelOrganizer.setup')(
        docstrings.dedent(ModelOrganizer.setup))

    @docstrings.dedent
    def setup(self, root_dir, projectname=None, link=False, src_project=None,
              compiler=None, **kwargs):
        """
        Perform the initial setup for the model

        Parameters
        ----------
        %(ModelOrganizer.setup.parameters)s
        link: bool
            If set, the source files are linked to the original ones instead
            of copied
        src_project: str
            Another model name to use the source model files from
        compiler: str
            The path to the compiler to use. If None, the global compiler
            option is used
        """
        root_dir = super(GWGENOrganizer, self).setup(
            root_dir, projectname=projectname, **kwargs)
        self.config.projects[self.projectname]['src'] = src_dir = osp.join(
            root_dir, 'src')
        if not osp.exists(src_dir):
            os.makedirs(src_dir)
        if src_project:
            module_src = self.config.projects[src_project]['src']
        else:
            module_src = osp.join(osp.dirname(__file__), 'src')
        for f in os.listdir(module_src):
            target = osp.join(src_dir, f)
            if osp.exists(target):
                os.remove(target)
            if link:
                self._link(osp.join(module_src, f), target)
            else:
                shutil.copyfile(osp.join(module_src, f), target)
        compiler = compiler or self.global_config.get('compiler')
        if compiler is not None:
            with open(osp.join(src_dir, 'Makefile')) as f:
                make_file = f.read()
            make_file = re.sub('^\s*FC\s*=\s*.*$', 'FC = ' + compiler,
                               make_file, flags=re.MULTILINE)
            with open(osp.join(src_dir, 'Makefile'), 'w') as f:
                f.write(make_file)
        return root_dir

    def _modify_setup(self, parser):
        parser.setup_args(ModelOrganizer.setup)
        self._modify_app_main(parser)
        parser.update_arg('src_project', short='src')
        parser.update_arg('compiler', short='c')

    @docstrings.dedent
    def compile_model(self, projectname=None, **kwargs):
        """
        Compile the model

        Parameters
        ----------
        projectname: str
            The name of the project. If None, use the last one or the one
            specified by the current experiment
        ``**kwargs``
            Keyword arguments passed to the :meth:`app_main` method
        """
        import subprocess as spr
        self.app_main(**kwargs)
        projectname = projectname or self.projectname
        self.projectname = projectname
        self.logger.info("Compiling %s", projectname)
        pdict = self.config.projects[projectname]
        pdict['bindir'] = bin_dir = osp.join(pdict['root'], 'bin')
        pdict['bin'] = osp.join(bin_dir, 'weathergen')
        src_dir = self.abspath(pdict['src'])
        if not os.path.exists(bin_dir):
            self.logger.debug("    Creating bin directory %s", bin_dir)
            os.makedirs(bin_dir)
        for f in os.listdir(src_dir):
            self.logger.debug("    Linking %s...", f)
            target = osp.join(bin_dir, f)
            if osp.exists(target):
                os.remove(target)
            self._link(osp.join(src_dir, f), target)
        spr.check_call(['make', '-C', bin_dir, 'all'], stdout=sys.stdout,
                       stderr=sys.stderr)
        self.logger.debug('Compilation done.')
        ts = self.project_config['timestamps']
        ts['compile'] = ts['compile_model '] = dt.datetime.now()

    def _modify_compile_model(self, parser):
        """Does nothing since compile takes no special arguments"""
        self._modify_app_main(parser)

    # -------------------------------------------------------------------------
    # -------------------------- Configuration --------------------------------
    # ------------------ Parts for configuring the organizer ------------------
    # -------------------------------------------------------------------------

    docstrings.get_sectionsf("ModelOrganizer.configure")(docstrings.dedent(
        ModelOrganizer.configure))

    @docstrings.dedent
    def configure(self, update_nml=None, max_stations=None,
                  datadir=None, database=None, user=None, host=None, port=None,
                  chunksize=None, compiler=None, **kwargs):
        """
        Configure the projects and experiments

        Parameters
        ----------
        %(ModelOrganizer.configure.parameters)s
        update_nml: str or dict
            A python dict or path to a namelist to use for updating the
            namelist of the experiment
        max_stations: int
            The maximum number of stations to process in one parameterization
            process. Does automatically impact global settings
        datadir: str
            Path to the data directory to use (impacts the project
            configuration)
        database: str
            The name of a postgres data base to write the data to
        user: str
            The username to use when logging into the database
        host: str
            the host which runs the database server
        port: int
            The port to use to log into the the database
        chunksize: int
            The chunksize to use for the parameterization and evaluation
        compiler: str
            The path to the fortran compiler to use"""
        super(GWGENOrganizer, self).configure(**kwargs)
        exp_config = self.exp_config
        if update_nml is not None:
            import f90nml
            with open(update_nml) as f:
                ref_nml = f90nml.read(f)
            nml2use = exp_config.setdefault('namelist', OrderedDict())
            for key, nml in ref_nml.items():
                nml2use.setdefault(key, OrderedDict()).update(dict(nml))
        gconf = self.config.global_config
        if max_stations:
            gconf['max_stations'] = max_stations
        if datadir:
            datadir = osp.abspath(datadir)
            self.project_config['data'] = datadir
        if database is not None:
            exp_config['database'] = database
        if user is not None:
            gconf['user'] = user
        if port is not None:
            gconf['port'] = port
        if host is not None:
            gconf['host'] = '127.0.0.1'
        if chunksize is not None:
            gconf['chunksize'] = chunksize
        if compiler is not None:
            gconf['compiler'] = compiler

    def _modify_configure(self, parser):
        parser.setup_args(super(GWGENOrganizer, self).configure)
        super(GWGENOrganizer, self)._modify_configure(parser)
        parser.update_arg('datadir', short='d')
        parser.update_arg('update_nml', short='u')
        parser.update_arg('max_stations', short='max', type=int)
        parser.update_arg('database', short='db')
        parser.update_arg('compiler', short='c')

    # -------------------------------------------------------------------------
    # -------------------------- Preprocessing --------------------------------
    # -------------- Preprocessing functions for the experiment ---------------
    # -------------------------------------------------------------------------

    @property
    def preproc_funcs(self):
        """A mapping from preproc commands to the corresponding function"""
        return {'select': self.select,
                'cloud': self.cloud_preproc,
                'test': self.create_test_sample}

    @docstrings.dedent
    def preproc(self, **kwargs):
        """
        Preprocess the data

        Parameters
        ----------
        ``**kwargs``
            Any keyword from the :attr:`preproc` attribute with kws for the
            corresponding function, or any keyword for the :meth:`main` method
        """
        funcs = self.preproc_funcs
        sp_kws = {key: kwargs.pop(key) for key in set(kwargs).intersection(
            funcs)}
        self.app_main(**kwargs)
        exp_config = self.fix_paths(self.exp_config)
        outdir = exp_config.setdefault('indir', osp.join(
            exp_config['expdir'], 'input'))
        if not osp.exists(outdir):
            os.makedirs(outdir)

        preproc_config = exp_config.setdefault('preproc', OrderedDict())

        for key, val in sp_kws.items():
            if isinstance(val, Namespace):
                val = vars(val)
            info = funcs[key](**val)
            if info:
                preproc_config[key] = info

    def _modify_preproc(self, parser):
        from gwgen.preproc import CloudPreproc
        self._modify_app_main(parser)
        sps = parser.add_subparsers(title='Preprocessing tasks', chain=True)

        # select
        sp = sps.add_parser(
            'select', help='Select stations based upon a regular grid')
        sp.setup_args(self.select)
        sp.update_arg('grid', short='g')
        sp.update_arg('grid_output', short='og')
        sp.update_arg('stations_output', short='os')
        sp.update_arg('igrid_key', short='k')
        sp.update_arg('grid_key', short='ok')
        sp.update_arg('grid_db', short='gdb')
        sp.update_arg('stations_db', short='sdb')
        sp.update_arg('no_prcp_check', short='nc')
        sp.update_arg('setup_from', short='f', long='from',
                      dest='setup_from')
        sp.update_arg('download', short='d', choices=['single', 'all'])

        # cloud preprocessing
        sp = sps.add_parser('cloud', help='Cloud preprocessing')
        sp.setup_args(self.cloud_preproc)
        sp.update_arg('max_files', short='mf', type=int)
        sp.pop_arg('return_manager')
        self._modify_task_parser(sp, CloudPreproc)

        # test samples
        sp = sps.add_parser(
            'test', help='Create a test sample for selected GHCN stations')
        sp.setup_args(self.create_test_sample)
        sp.update_arg('no_cloud', short='nc')
        sp.update_arg('reduce_eecra', short='re', type=float)
        sp.update_arg('keep_all', short='a')

        return parser

    # ------------------------------- Selection -------------------------------

    def _prcp_check(self, series):
        try:
            return 11 == len(series.to_frame().set_index('prcp').join(
                self._prcp_test, how='inner').prcp.unique())
        except:
            return None

    def _select_best_df(self, df, test_series, kws):
        from gwgen.parameterization import DailyGHCNData
        # disable logging for the DailyGHCNData task
        task_logger = DailyGHCNData([], self.exp_config, self.project_config,
                                    self.global_config).logger
        orig_level = task_logger.level
        task_logger.setLevel(logging.WARNING)
        self._test_series = test_series
        self._select_kws = kws
        self._select_task = DailyGHCNData
        g = df.sort_values('nyrs', ascending=False).groupby(
            level=['clon', 'clat'])
        ret = g.id.agg(self._select_best)
        task_logger.setLevel(orig_level)
        return ret

    def _select_best(self, series):
        test_series = self._test_series
        for station in series.values:
            task = self._select_task(
                np.array([station]), self.exp_config, self.project_config,
                self.global_config, **self._select_kws)
            try:
                task.init_task()
            except FileNotFoundError as e:
                task.logger.warn(e)
            else:
                task.setup()
                if len(test_series) == len(
                    task.data.set_index('prcp').join(
                        test_series, how='inner').prcp.unique()):
                    return station
        return series.values[0]

    @staticmethod
    def _parallel_select(l):
        organizer, df, test_series, kws = l
        return organizer._select_best_df(df, test_series, kws)

    @docstrings.dedent
    def select(self, grid=None, grid_output=None, stations_output=None,
               igrid_key=None, grid_key=None, grid_db=None, stations_db=None,
               no_prcp_check=False, setup_from=None, download=None, **kwargs):
        """
        Select stations based upon a regular grid

        Parameters
        ----------
        grid: str
            The path to a csv-file containing a lat and a lon column with the
            information on the centers of the grid. If None, `igrid_key` must
            not be None and point to a key in the configuration (either the one
            of the experiment, or the project, or the global configuration)
            specifying the path
        grid_output: str
            The path to the csv-file where to store the mapping from grid
            lat-lon to station id.
        stations_output: str
            The path to the csv-file where to store the mapping from station
            to grid center point
        igrid_key: str
            The key in the configuration where to store the path of the `grid`
            input file
        grid_key: str
            The key in the configuration where to store the name of the
            `grid_output` file.
        grid_db: str
            The name of a data table to store the data of `stations_output` in
        stations_db: str
            The name of a data table to store the data for `stations_output` in
        no_prcp_check: bool
            If True, we will not check for the values between 0.1 and 1.0 for
            precipitation and save the result in the ``'best'`` column
        setup_from: { 'scratch' | 'file' | 'db' }
            The setup method for the daily data for the prcp check
        download: { 'single' | 'all' }
            Handles how to manage missing files for the prcp check. If None
            (default), an warning is printed and the file is ignored, if
            ``'single'``, the missing file is downloaded, if ``'all'``, the
            entire tarball is downloaded (strongly not recommended for this
            function)

        Other Parameters
        ----------------
        ``**kwargs``
            are passed to the :meth:`main` method

        Notes
        -----
        for `igrid_key` and `ogrid_key` we recommend one of
        ``{'grid', 'param_grid', 'eval_grid'`` because that implies a
        correct path management
        """
        from gwgen.evaluation import EvaluationPreparation
        import numpy as np
        import scipy.spatial
        import pandas as pd

        logger = self.logger

        if grid is None:
            if igrid_key is not None:
                grid = self.exp_config.get(igrid_key, self.project_config.get(
                    igrid_key, self.global_config.get(igrid_key)))
            else:
                raise ValueError(
                    "No grid file or configuration key specified!")
        if grid is None:
            raise ValueError(
                    "No grid file specified and '%s' could not be found in "
                    "the configuration!" % igrid_key)
        t = EvaluationPreparation(np.array([]), self.exp_config,
                                  self.project_config, self.global_config)
        # get inventory
        t.download_src()
        df_stations = t.station_list
        df_stations = df_stations[df_stations.vname == 'PRCP'].drop(
            'vname', 1).reset_index()  # reset_index required due to filtering
        df_stations['nyrs'] = df_stations.lastyr - df_stations.firstyr

        # read 1D grid information
        df_centers = pd.read_csv(grid)
        df_centers.rename(columns={'lon': 'clon', 'lat': 'clat'}, inplace=True)

        # concatenate lat and lon values into x-y points
        center_points = np.dstack(
            [df_centers.clat.values, df_centers.clon.values])[0]
        station_points = np.dstack([df_stations.lat, df_stations.lon])[0]

        # look up the nearest neighbor
        logger.debug('Searching neighbors...')
        kdtree = scipy.spatial.cKDTree(center_points)
        dist, indexes = kdtree.query(station_points)
        logger.debug('Done.')

        # store the lat and longitude of, and the distance to the center grid
        # point in the stations table
        df_stations['clon'] = df_centers.clon.values[indexes]
        df_stations['clat'] = df_centers.clat.values[indexes]
        df_stations['dist'] = dist

        # --------- stations with the closest distance to grid center ---------
        # group by the center coordinates and look for the index with the
        # smallest distance
        g = df_stations.sort_index().groupby(['clon', 'clat'])
        indices_closest = g.dist.idxmin()
        indices_longest = g.nyrs.idxmax()
        # merge the nearest stations into the centers table
        df_centers.set_index(['clon', 'clat'], inplace=True)
        df_stations.set_index(['clon', 'clat'], inplace=True)
        merged = df_centers.merge(
            df_stations.ix[indices_closest][['id']].rename(
                columns={'id': 'nearest_station'}),
            left_index=True, right_index=True, how='outer')
        merged = merged.merge(
            df_stations.ix[indices_longest][['id']].rename(
                columns={'id': 'longest_record'}),
            left_index=True, right_index=True, how='outer')

        if not no_prcp_check:
            test_series = pd.Series(
                np.arange(0.1, 1.05, 0.1), name='prcp')
            logger.debug('Performing best station check with %s',
                         test_series.values)
            kws = dict(download=download, setup_from=setup_from)
            if not self.global_config.get('serial'):
                import multiprocessing as mp
                nprocs = self.global_config.get('nprocs', 'all')
                lonlats = np.unique(df_stations.dropna(0).index.values)
                if nprocs == 'all':
                    nprocs = mp.cpu_count()
                splitted = np.array_split(lonlats, nprocs)
                try:
                    nprocs = list(map(len, splitted)).index(0)
                except ValueError:
                    pass
                else:
                    splitted = splitted[:nprocs]
                dfs = [df_stations.loc[list(arr)] for arr in splitted]
                # initializing pool
                logger.debug('Start %i processes', nprocs)
                pool = mp.Pool(nprocs)
                args = list(zip(repeat(self), dfs, repeat(test_series),
                                repeat(kws)))
                res = pool.map_async(self._parallel_select, args)
                best = pd.concat(res.get())
                pool.close()
                pool.join()
                pool.terminate()
            else:
                best = self._select_best_df(
                    df_stations.dropna(0), test_series, kws)
            merged = merged.merge(
                best.to_frame().rename(columns={'id': 'best'}),
                left_index=True, right_index=True, how='outer')

        if igrid_key:
            self.exp_config[igrid_key] = grid
        if stations_output:
            logger.debug('Dumping to%s %s',
                         ' exisiting' if osp.exists(stations_output) else '',
                         stations_output)
            utils.safe_csv_append(df_stations, stations_output)

        if grid_output:
            logger.debug('Dumping to%s %s',
                         ' exisiting' if osp.exists(grid_output) else '',
                         grid_output)
            utils.safe_csv_append(merged, grid_output)
            if grid_key is not None:
                self.exp_config[grid_key] = grid_output
        if stations_db or grid_db:
            conn = t.engine.connect()
            if stations_db:
                logger.info('Writing %i lines into %s', len(df_stations),
                            stations_db)
                df_stations.to_sql(stations_db, conn, if_exists='append')
            if grid_db:
                logger.info('Writing %i lines into %s', len(merged),
                            grid_db)
                merged.to_sql(grid_db, conn, if_exists='append')
            conn.close()

        return df_stations, merged

    # --------------------------- Cloud inventory -----------------------------

    @docstrings.dedent
    def cloud_preproc(self, max_files=None, return_manager=False, **kwargs):
        """
        Extract the inventory of EECRA stations

        Parameters
        ----------
        max_files: int
            The maximum number of files to process during one process. If None,
            it is determined by the global ``'max_stations'`` key
        ``**kwargs``
            Any task in the :class:`gwgen.preproc.CloudPreproc` framework
        """
        from gwgen.preproc import CloudPreproc
        from gwgen.parameterization import HourlyCloud
        stations_orig = self.global_config.get('max_stations')
        if max_files is not None:
            self.global_config['max_stations'] = max_files
        files = HourlyCloud.from_organizer(self, []).raw_src_files
        manager = CloudPreproc.get_manager(config=self.global_config)
        for key, val in kwargs.items():
            if isinstance(val, Namespace):
                kwargs[key] = val = vars(val)
                val.pop('max_files', None)
        self._setup_manager(manager, stations=list(files.values()),
                            base_kws=kwargs)
        d = {}
        manager.run(d)
        if stations_orig:
            self.global_config['max_stations'] = stations_orig
        else:
            self.global_config.pop('max_stations', None)
        if return_manager:
            return d, manager
        else:
            return d

    # --------------------------- Parameterization ----------------------------

    @docstrings.get_sectionsf('GWGENOrganizer.param')
    @docstrings.dedent
    def param(self, complete=False, stations=None, other_exp=None,
              setup_from=None, to_db=None, to_csv=None, database=None,
              norun=False, to_return=None, **kwargs):
        """
        Parameterize the experiment

        Parameters
        ----------
        stations: str or list of str
            either a list of stations to use or a filename containing a
            1-row table with stations
        other_exp: str
            Use the configuration from another experiment
        setup_from: str
            Determine where to get the data from. If `scratch`, the
            data will be calculated from the raw data. If `file`,
            the data will be loaded from a file, if `db`, the data
            will be loaded from a postgres database (Note that the
            `database` argument must be provided!).
        to_db: bool
            Save the data into a postgresql database (Note that the
            `database` argument must be provided!)
        to_csv: bool
            Save the data into a csv file
        database: str
            The name of a postgres data base to write the data to
        norun: bool, list of str or ``'all'``
            If True, only the data is set up and the configuration of the
            experiment is not affected. It can be either a list of  tasks or
            True or ``'all'``
        to_return: list of str or ``'all'``
            The names of the tasks to return. If None, only the ones with an
            :attr:`gwgen.utils.TaskBase.has_run` are returned.
        complete: bool
            If True, setup and run all possible tasks
        """
        from gwgen.parameterization import Parameterizer
        task_names = [task.name for task in Parameterizer._registry]
        parameterizer_kws = {
            key: vars(val) if isinstance(val, Namespace) else dict(val)
            for key, val in kwargs.items() if key in task_names}
        main_kws = {key: val for key, val in kwargs.items()
                    if key not in task_names}
        self.app_main(**main_kws)
        experiment = self.experiment
        exp_dict = self.fix_paths(self.config.experiments[experiment])
        param_dir = exp_dict.setdefault(
            'paramdir', osp.join(exp_dict['expdir'], 'parameterization'))
        if not osp.exists(param_dir):
            os.makedirs(param_dir)
        projectname = self.projectname
        logger = self.logger
        logger.info("Parameterizing experiment %s of project %s",
                    experiment, projectname)
        stations = self._get_stations(stations, other_exp, param_dir,
                                      'param_stations')
        global_conf = self.config.global_config
        # choose keywords for data processing
        manager = Parameterizer.get_manager(config=global_conf)
        self._setup_manager(manager, stations, other_exp, setup_from, to_db,
                            to_csv, database, to_return, complete,
                            parameterizer_kws)
        # update experiment namelist and configuration
        if not norun:
            manager.run(exp_dict.setdefault('parameterization', OrderedDict()),
                        exp_dict.setdefault('namelist', OrderedDict()))
        return manager

    def _modify_param(self, parser, *args, **kwargs):
        from gwgen.parameterization import Parameterizer
        self._modify_task_parser(parser, Parameterizer, *args, **kwargs)

    # --------------------------------- Test ----------------------------------

    @docstrings.dedent
    def create_test_sample(self, test_dir, stations, no_cloud=False,
                           reduce_eecra=0, keep_all=False):
        """
        Create a test sample for the given GHCN stations

        Parameters
        ----------
        test_dir: str
            The path to the directory containing the test files from Github
        stations: str or list of str
            either a list of GHCN stations to use or a filename containing a
            1-row table with GHCN stations
        no_cloud: bool
            If True, no cloud stations are extracted
        reduce_eecra: float
            The percentage by which to reduce the EECRA data
        keep_all: bool
            If True all years of the EECRA data are used. Otherwise, only the
            years with complete temperature and cloud are kept. Note
            that this has only an effect if `reduce_eecra` is not 0
        """
        import calendar
        import pandas as pd
        from gwgen.parameterization import DailyGHCNData, HourlyCloud

        def is_complete(s):
            ndays = 366 if calendar.isleap(s.name[1]) else 365
            s[:] = s.ix[~s.index.duplicated()].count() == ndays
            return s

        stations = self._get_stations(stations)
        np.savetxt(osp.join(test_dir, 'test_stations.dat'), stations, fmt='%s')
        # download the GHCN data
        ghcn_task = DailyGHCNData.from_organizer(self, stations,
                                                 download='single')
        ghcn_task.init_from_scratch()
        data_dir = super(DailyGHCNData, ghcn_task).data_dir

        if not no_cloud:
            eecra_task = HourlyCloud.from_organizer(self, stations)
            if len(eecra_task.stations) == 0:
                raise ValueError(
                    "Could not find any station in the given stations %s!",
                    ', '.join(stations))
            np.savetxt(osp.join(test_dir, 'eecra_test_stations.dat'),
                       eecra_task.eecra_stations, fmt='%i')
            eecra_task.init_from_scratch()

        for fname in ghcn_task.raw_src_files:
            target = fname.replace(osp.join(data_dir, ''),
                                   osp.join(test_dir, ''))
            if not osp.samefile(fname, target):
                shutil.copyfile(fname, target)
            shutil.make_archive(osp.join(test_dir, 'ghcn', 'ghcnd_all'),
                                'gztar',
                                root_dir=osp.join(test_dir, 'ghcn'),
                                base_dir='ghcnd_all')

        if not no_cloud:
            for fname in eecra_task.src_files:
                target = fname.replace(osp.join(data_dir, ''),
                                       osp.join(test_dir, ''))
                if not reduce_eecra and not osp.samefile(fname, target):
                    shutil.copyfile(fname, target)
                else:
                    df = pd.read_csv(fname)
                    if not keep_all:
                        df_bool = df.set_index(
                            ['station_id', 'year', 'month', 'day'])[[
                                'ww', 'AT', 'N']]
                        for col in df_bool.columns:
                            df_bool[col] = df_bool[col].astype(bool)
                        g = df_bool.groupby(level=['station_id', 'year'])
                        mask = g.transform(is_complete).values.any(axis=1)
                        df = df.ix[mask]

                    g = df.groupby(['station_id', 'year'],
                                   as_index=False)
                    tot = g.ngroups
                    n = np.ceil(tot * (100 - reduce_eecra) / 100)
                    idx_groups = iter(sorted(np.random.permutation(tot)[:n]))
                    self.logger.debug(
                        'Saving EECRA test sample with %i years from %i to '
                        '%s', n, tot, target)
                    df.ix[1:0].to_csv(target, index=False)
                    igrp = next(idx_groups)
                    for i, (key, group) in enumerate(g):
                        if i == igrp:
                            group.to_csv(target, header=False, mode='a',
                                         index=False)
                            igrp = next(idx_groups, -1)
    # -------------------------------------------------------------------------
    # ------------------------------- Run -------------------------------------
    # --------------------------- Run the experiment --------------------------
    # -------------------------------------------------------------------------

    @docstrings.get_sectionsf('GWGENOrganizer.run')
    @docstrings.dedent
    def run(self, ifile=None, ofile=None, odir=None, work_dir=None,
            remove=False, **kwargs):
        """
        Run the experiment

        Parameters
        ----------
        ifile: str
            The path to the input file. If None, it is assumed that it is
            stored in the ``'input'`` key in the experiment configuration
        ofile: str
            The path to the output file.  If None, it is assumed that it is
            stored in the ``'input'`` key in the experiment configuration or
            it will be stored in ``'odir/exp_id.csv'``. The output directory
            ``'odir'`` is determined by the `odir` parameter
        odir: str
            The path to the output directory. If None and not already saved
            in the configuration, it will default to
            ``'experiment_dir/outdata'``
        work_dir: str
            The path to the work directory where the binaries are copied to.
        remove: bool
            If True, the `work_dir` will be removed if it already exists

        Other Parameters
        ----------------
        ``**kwargs``
            Will be passed to the :meth:`main` method
        """
        import subprocess as spr
        import stat
        import f90nml
        from copy import deepcopy
        self.app_main(**kwargs)
        logger = self.logger
        exp_config = self.fix_paths(self.exp_config)
        project_config = self.fix_paths(self.project_config)
        experiment = self.experiment
        if not {'compile_model', 'compile'} & set(
                project_config['timestamps']):
            self.compile_model(**kwargs)
        logger.info("Running experiment %s of project %s",
                    experiment, self.projectname)
        if ifile is None:
            ifile = exp_config.get('input', self.project_config.get(
                'input',  self.global_config.get('input')))
        if ifile is None:
            raise ValueError("No input file specified!")
        if ofile is None:
            ofile = exp_config.get('outdata')
        if ofile is None:
            ofile = osp.join(
                odir or exp_config.get(
                    'outdir', osp.join(exp_config['expdir'], 'outdata')),
                str(experiment) + '.csv')
        if work_dir is None:
            work_dir = exp_config.get('workdir',
                                      osp.join(exp_config['expdir'], 'work'))
        exp_config['outdir'] = odir = osp.dirname(ofile)
        exp_config['outdata'] = ofile
        exp_config['input'] = ifile
        exp_config['indir'] = osp.dirname(ifile)
        exp_config['workdir'] = work_dir
        nml = exp_config.setdefault(
            'namelist', {'weathergen_ctl': OrderedDict(),
                         'main_ctl': OrderedDict()})
        for key in ['weathergen_ctl', 'main_ctl']:
            nml.setdefault(key, {})

        if osp.exists(work_dir) and remove:
            shutil.rmtree(work_dir)
        elif not osp.exists(work_dir):
            os.makedirs(work_dir)
        if not osp.exists(odir):
            os.makedirs(odir)

        f = project_config['bin']
        target = osp.join(work_dir, osp.basename(f))
        logger.debug('Copy executable %s to %s', f, target)
        shutil.copyfile(f, target)
        os.chmod(target, stat.S_IWUSR | stat.S_IXUSR | stat.S_IRUSR)
        logger.debug('    Name list: %s', ordered_yaml_dump(nml))
        nml = deepcopy(nml)
        # transpose multidimensional arrays because they get transposed by
        # f90nml. Otherwise you get errors using functions like matmul
        for key, sub_nml in nml.items():
            for key2, val in sub_nml.items():
                if np.ndim(val) >= 2:
                    sub_nml[key2] = np.round(np.transpose(val), 8).tolist()
        with open(osp.join(work_dir, 'weathergen.nml'), 'w') as f:
            f90nml.write(nml, f)

        logger.debug('Running experiment...')
        logger.debug('    input: %s', ifile)
        logger.debug('    output: %s', ofile)
        t = dt.datetime.now()
        commands = 'cd %s && %s %s %s' % (work_dir, target, ifile, ofile)
        logger.debug(commands)
        spr.check_call(commands, stdout=sys.stdout, stderr=sys.stderr,
                       shell=True)
        err_msg = "Failed to run the experiment with '%s'!" % commands
        if not osp.exists(ofile):
            raise RuntimeError(
                (err_msg + "Reason: Output %s missing" % (ofile)))
        else:  # check if the file contains more than one line
            with open(ofile) as f:
                f.readline()
                if f.tell() == os.fstat(f.fileno()).st_size:
                    raise RuntimeError(
                        (err_msg + "Reason: Output %s is empty" % (ofile)))
        logger.debug('Done. Time needed: %s', dt.datetime.now() - t)

    def _modify_run(self, parser):
        parser.update_arg('ifile', short='i')
        parser.update_arg('ofile', short='o')
        parser.update_arg('odir', short='od')
        parser.update_arg('work_dir', short='wd')
        parser.update_arg('remove', short='r')

    # -------------------------------------------------------------------------
    # -------------------------- Postprocessing -------------------------------
    # ------------ Postprocessing functions for the experiment ----------------
    # -------------------------------------------------------------------------

    # ---------------------------- Evaluation ---------------------------------

    @docstrings.get_sectionsf('GWGENOrganizer.evaluate')
    @docstrings.dedent
    def evaluate(self, stations=None, other_exp=None,
                 setup_from=None, to_db=None, to_csv=None, database=None,
                 norun=False, to_return=None, complete=False, **kwargs):
        """
        Evaluate the experiment

        Parameters
        ----------
        %(GWGENOrganizer.param.parameters)s"""
        from gwgen.evaluation import Evaluator
        task_names = [task.name for task in Evaluator._registry]
        evaluator_kws = {
            key: vars(val) if isinstance(val, Namespace) else dict(val)
            for key, val in kwargs.items() if key in task_names}
        main_kws = {key: val for key, val in kwargs.items()
                    if key not in task_names}
        self.app_main(**main_kws)
        experiment = self.experiment
        exp_dict = self.fix_paths(self.config.experiments[experiment])
        eval_dir = exp_dict.setdefault(
            'evaldir', osp.join(exp_dict['expdir'], 'evaluation'))
        if not osp.exists(eval_dir):
            os.makedirs(eval_dir)
        projectname = self.projectname
        logger = self.logger
        logger.info("Evaluating experiment %s of project %s",
                    experiment, projectname)
        stations = self._get_stations(stations, other_exp, eval_dir,
                                      'eval_stations')
        global_conf = self.config.global_config
        # choose keywords for data processing
        manager = Evaluator.get_manager(config=global_conf)
        self._setup_manager(manager, stations, other_exp, setup_from, to_db,
                            to_csv, database, to_return, complete,
                            evaluator_kws)
        # update experiment namelist and configuration
        if not norun:
            manager.run(exp_dict.setdefault('evaluation', OrderedDict()))
        return manager

    def _modify_evaluate(self, parser, *args, **kwargs):
        from gwgen.evaluation import Evaluator
        self._modify_task_parser(parser, Evaluator, *args, **kwargs)

    @property
    def bias_correction_methods(self):
        return {'wind': self.wind_bias_correction,
                'tmin': self.tmin_bias_correction}

    @docstrings.dedent
    def bias_correction(
            self, keep=False, quantiles=list(range(1, 100)),
            no_evaluation=False, new_project=False, **kwargs):
        """
        Perform a bias correction for the data

        Parameters
        ----------
        keep: bool
            If not True, the experiment configuration files are not modified.
            Otherwise the `quants` section is kept for the given quantiles
        quantiles: list of float
            The quantiles to use for the bias correction. Does not have an
            effect if `no_evaluation` is set to True
        no_evaluation: bool
            If True, the existing evaluation in the configuration is used for
            the bias correction
        new_project: bool
            If True, a new project will be created even if a file in
            `project_output` exists already

        Returns
        -------
        dict
            The results of the underlying bias correction methods"""
        methods = self.bias_correction_methods

        main_kws = self.get_app_main_kwargs(kwargs, keep=True)
        bias_kws = {
            key: kwargs.pop(key) for key in set(methods).intersection(kwargs)}

        self.app_main(**main_kws)
        self.logger.debug('Calculating bias correction for experiment %s',
                          self.experiment)
        old = self.exp_config.get('evaluation', {}).get('quants')
        postproc_dir = self.exp_config.setdefault(
            'postprocdir', osp.join(self.exp_config['expdir'], 'postproc'))
        if not osp.exists(postproc_dir):
            os.makedirs(postproc_dir)
        quants_output = osp.join(postproc_dir, 'quants_bias')
        kwargs['quants'] = {'quantiles': quantiles, 'transform_wind': False,
                            'new_project': new_project,
                            'names': list(bias_kws),
                            'project_output': quants_output + '.pkl',
                            'plot_output': quants_output + '.pdf',
                            'nc_output': quants_output + '.nc'}
        self.evaluate(**kwargs)

        d = self.exp_config.setdefault('postproc', OrderedDict()).setdefault(
            'bias', OrderedDict())

        d['plot_file'] = quants_output + '.pdf'
        d['project_file'] = quants_output + '.pkl'
        d['nc_file'] = quants_output + '.nc'

        for name, kws in bias_kws.items():
            if isinstance(kws, Namespace):
                kws = vars(kws)
                for key in ['keep', 'quantiles', 'no_evaluation']:
                    kws.pop(key, None)
            methods[name](self.exp_config['evaluation']['quants'], **kws)

        if not keep:
            if old:
                self.exp_config['evaluation']['quants'] = old
            else:
                self.exp_config['evaluation'].pop('quants')

    def _modify_bias_correction(self, parser):
        self._modify_app_main(parser)
        parser.update_arg('keep', short='k')
        parser.update_arg(
            'quantiles', short='q', type=utils.str_ranges,
            metavar='f1[,f21[-f22[-f23]]]', help=docstrings.dedents("""
                The quantiles to use for calculating the percentiles.
                %(str_ranges.s_help)s."""))
        parser.pop_key('quantiles', 'nargs', None)
        parser.update_arg('new_project', short='np')
        parser.update_arg('no_evaluation', short='ne')

        sps = parser.add_subparsers(chain=True)

        # -- wind
        sp = sps.add_parser('wind')
        sp.setup_args(self.wind_bias_correction_logistic)
        sp.setup_args(self.wind_bias_correction)
        sp.update_arg('new_project', short='np')
        sp.update_arg('plot_output', short='po')
        sp.pop_arg('info')
        sp.pop_arg('close')

        # -- tmin
        sp = sps.add_parser('tmin')
        sp.setup_args(self.poly_bias_correction)
        sp.pop_arg('vname')
        sp.pop_arg('what')
        sp.pop_arg('ds')
        sp.setup_args(self.tmin_bias_correction)
        sp.update_arg('new_project', short='np')
        sp.update_arg('plot_output', short='po')
        sp.pop_arg('info')
        sp.pop_arg('close')

    @docstrings.get_sectionsf('GWGENOrganizer.wind_bias_correction_logistic')
    @docstrings.dedent
    def wind_bias_correction_logistic(
            self, info, new_project=False, plot_output=None, close=True):
        """
        Perform a bias correction for the data

        Parameters
        ----------
        info: dict
            The configuration of the quantile evaluation
        new_project: bool
            If True, a new project will be created even if a file in
            `project_output` exists already
        plot_output: str
            The name of the output file. If not specified, it defaults to
            `<exp_dir>/postproc/<vname>_bias_correction.pdf`
        close: bool
            If True, close the project at the end"""
        import pandas as pd
        from scipy import stats
        import xarray as xr
        import psyplot.project as psy

        vname = 'wind'

        self.logger.debug('Calculating bias correction for experiment %s',
                          self.experiment)
        postproc_dir = self.exp_config.setdefault(
            'postprocdir', osp.join(self.exp_config['expdir'], 'postproc'))

        df = pd.DataFrame.from_dict(info[vname], 'index')
        try:
            # drop all percentiles
            df.drop('All', inplace=True)
        except (ValueError, KeyError) as e:
            pass
        df.index.name = 'pctl'
        df.reset_index(inplace=True)
        df['unorm'] = stats.norm.ppf(
            df['pctl'].astype(float) / 100., 0, 1.0)
        ds = xr.Dataset.from_dataframe(df)

        # --- plots
        d = self.exp_config.setdefault('postproc', OrderedDict()).setdefault(
            'bias', OrderedDict()).setdefault(vname, OrderedDict())
        plot_output = plot_output or d.get('plot_output')
        if plot_output is None:
            plot_output = osp.join(
                postproc_dir, vname + '_bias_correction.pdf')

        project_output = osp.splitext(plot_output)[0] + '.pkl'
        nc_output = osp.splitext(plot_output)[0] + '.nc'

        d['plot_file'] = plot_output
        d['project_file'] = project_output
        d['nc_file'] = nc_output

        # --- slope bias correction
        if osp.exists(project_output) and not new_project:
            mp = psy.Project.load_project(project_output, datasets=[ds])
            sp2 = mp.linreg(name='slope')
        else:
            import seaborn as sns
            sns.set_style('white')
            sp1 = psy.plot.lineplot(ds, name='slope', coord='unorm',
                                    linewidth=0, marker='o', legend=False)
            sp2 = psy.plot.linreg(
                 ds, name='slope', ax=sp1[0].psy.ax,
                 coord='unorm', fit=logistic_function,
                 ylabel=('$\\mathrm{{Simulated}}\\, \\mathrm{{%s}} / '
                         '\\mathrm{{Observed}}\\, \\mathrm{{%s}}$') % (
                            vname, vname),
                 legendlabels=(
                     '$\\frac{{\\mathrm{{Simulated}}}}'
                     '{{\\mathrm{{Observed}}}} = '
                     '\\frac{{%(L)4.3f}}{{1 + \\mathrm{{e}}^{{'
                     '%(k)4.3f\\cdot(x %(x0)+4.3f)}}}}$'),
                 legend={'fontsize': 'x-large', 'loc': 'upper left'},
                 xlabel='Random number $x$ from normal distribution')
            sp2.share(sp1[0], ['color', 'xlim', 'ylim'])
        arr = sp2.plotters[0].plot_data[0]
        nml = self.exp_config['namelist']['weathergen_ctl']
        if 'L' in arr.attrs:
            nml.pop(vname + '_bias_coeffs', None)
            for letter in ['L', 'k', 'x0']:
                nml[vname + '_slope_bias_' + letter] = float(arr.attrs[letter])
        else:  # polynomial fit
            for letter in ['L', 'k', 'x0']:
                nml.pop(vname + '_slope_bias_' + letter, None)
            nml[vname + '_bias_coeffs'] = [
                    float(arr.attrs.get('c%i' % i, 0.0)) for i in range(6)]

        # --- intercept bias correction
        if osp.exists(project_output) and not new_project:
            sp2 = mp.linreg(name='intercept')
        else:
            sp1 = psy.plot.lineplot(ds, name='intercept', coord='unorm',
                                    linewidth=0, marker='o', legend=False)
            sp2 = psy.plot.linreg(
                 ds, name='intercept', ax=sp1[0].psy.ax,
                 coord='unorm', fit=exponential_function,
                 ylabel=(
                    '$\\mathrm{{Simulated}}\\, \\mathrm{{%s}} - '
                    '\\mathrm{{Observed}}\\, \\mathrm{{%s}}$ [m/s]') % (
                        vname, vname),
                 legendlabels=(
                     '$\\mathrm{{Simulated}} - \\mathrm{{Observed}} ='
                     'e^{{%(a)1.4f \\cdot x %(b)+1.4f}}$'),
                 legend={'fontsize': 'medium', 'loc': 'upper left'},
                 xlabel='Random number $x$ from normal distribution')
        arr = sp2.plotters[0].plot_data[0]
        if 'a' in arr.attrs:
            nml.pop(vname + '_intercept_bias_coeffs', None)
            for letter in ['a', 'b']:
                nml[vname + '_intercept_bias_' + letter] = float(
                    arr.attrs[letter])
        else:  # polynomial fit
            for letter in ['a', 'b']:
                nml.pop(vname + '_intercept_bias_' + letter, None)
            nml[vname + '_intercept_bias_coeffs'] = [
                float(arr.attrs.get('c%i' % i, 0.0)) for i in range(6)]
        nml[vname + '_bias_min'] = float(ds.unorm.min().values)
        nml[vname + '_bias_max'] = float(ds.unorm.max().values)
        # --- save the data
        self.logger.info('Saving plots to %s', plot_output)
        mp = psy.gcp(True)
        mp.export(plot_output)
        self.logger.info('Saving project to %s', project_output)
        mp.save_project(project_output, paths=[nc_output])

        if close:
            psy.gcp(True).close(True, True, True)

    @docstrings.get_sectionsf('GWGENOrganizer.poly_bias_correction')
    @docstrings.dedent
    def poly_bias_correction(
            self, vname, what, info, new_project=False, plot_output=None,
            deg=3, close=True, ds=None):
        """
        Perform a bias correction based on percentile and a polynomial fit

        Parameters
        ----------
        vname: str
            The variable name to use
        what: str { 'slope' | 'intercept' }
            Either slope or intercept. The parameter that should be used for
            the bias correction
        info: dict
            The configuration of the quantile evaluation
        new_project: bool
            If True, a new project will be created even if a file in
            `project_output` exists already
        plot_output: str
            The name of the output file. If not specified, it defaults to
            `<exp_dir>/postproc/<vname>_bias_correction.pdf`
        deg: int
            The degree of the fittet polynomial
        close: bool
            If True, close the project at the end
        ds: xr.Dataset
            The xarray dataset to use. Otherwise it will be created from `info`
        """
        import pandas as pd
        from scipy import stats
        import xarray as xr
        import psyplot.project as psy

        def get_symbol(i):
            if not i:
                return ''
            elif i == 1:
                return 'x'
            else:
                return 'x^' + str(i)

        self.logger.debug('Calculating %s bias correction for experiment %s',
                          vname, self.experiment)
        postproc_dir = self.exp_config.setdefault(
            'postprocdir', osp.join(self.exp_config['expdir'], 'postproc'))
        if ds is None:
            df = pd.DataFrame(info[vname]).T
            try:
                # drop all percentiles
                df.drop('All', inplace=True)
            except (ValueError, KeyError) as e:
                pass
            df.index.name = 'pctl'
            df.reset_index(inplace=True)
            df['unorm'] = stats.norm.ppf(
                df['pctl'].astype(float) / 100., 0, 1.0)
            ds = xr.Dataset.from_dataframe(df)

        # --- plots
        d = self.exp_config.setdefault('postproc', OrderedDict()).setdefault(
            'bias', OrderedDict()).setdefault(vname, OrderedDict())
        plot_output = plot_output or d.get('plot_output')
        if plot_output is None:
            plot_output = osp.join(
                postproc_dir, vname + '_bias_correction.pdf')

        project_output = osp.splitext(plot_output)[0] + '.pkl'
        nc_output = osp.splitext(plot_output)[0] + '.nc'

        d['plot_file'] = plot_output
        d['project_file'] = project_output
        d['nc_file'] = nc_output

        if what == 'slope':
            ylabel = 'Simulated/Observed'
            if vname == 'wind':
                ylabel = '$\\sqrt{{' + ylabel + '}}$'
        else:
            ylabel = 'Simulated - Observed'
        diff_symbol = ylabel
        if vname == 'tmin':
            ylabel += ' [$^\circ$C]'

        # --- slope bias correction
        if osp.exists(project_output) and not new_project:
            mp = psy.Project.load_project(project_output, datasets=[ds])
            sp2 = mp.linreg
        else:
            import seaborn as sns
            sns.set_style('white')
            sp1 = psy.plot.lineplot(ds, name=what, coord='unorm',
                                    linewidth=0, marker='o', legend=False)
            label = '$%s = %s$' % (diff_symbol, ' '.join(
                '%(c{})+4.3f{}'.format(i, get_symbol(i))
                for i in range(deg + 1)))
            sp2 = psy.plot.linreg(
                 ds, name=what, ax=sp1[0].ax,
                 coord='unorm', fit='poly' + str(int(deg)),
                 ylabel=ylabel,
                 legendlabels=label,
                 legend={'fontsize': 'large', 'loc': 'upper left'},
                 xlabel='Random number from normal distribution')
            sp2.share(sp1[0], ['color', 'xlim', 'ylim'])
        attrs = sp2.plotters[0].plot_data[0].attrs
        nml = self.exp_config['namelist']['weathergen_ctl']
        nml[vname + '_bias_coeffs'] = [
            float(attrs.get('c%i' % i, 0.0)) for i in range(6)]
        nml[vname + '_bias_min'] = float(ds.unorm.min().values)
        nml[vname + '_bias_max'] = float(ds.unorm.max().values)

        # --- save the data
        self.logger.info('Saving plots to %s', plot_output)
        mp = psy.gcp(True)
        mp.export(plot_output)
        self.logger.info('Saving project to %s', project_output)
        mp.save_project(project_output, paths=[nc_output])

        if close:
            psy.gcp(True).close(True, True, True)

    docstrings.delete_params('GWGENOrganizer.poly_bias_correction.parameters',
                             'vname', 'what')

    @docstrings.dedent
    def tmin_bias_correction(self, *args, **kwargs):
        """
        Perform a bias correction for the minimum temperature data

        Parameters
        ----------
        %(GWGENOrganizer.poly_bias_correction.parameters.no_vname|what)s"""
        return self.poly_bias_correction('tmin', 'intercept', *args, **kwargs)

    @docstrings.dedent
    def wind_bias_correction(self, *args, **kwargs):
        """
        Perform a bias correction for the wind speed

        Parameters
        ----------
        %(GWGENOrganizer.wind_bias_correction_logistic.parameters)s"""
        return self.wind_bias_correction_logistic(*args, **kwargs)

    # ----------------------- Sensitivity analysis ----------------------------

    @docstrings.dedent
    def sensitivity_analysis(self, **kwargs):
        """
        Perform a sensitivity analysis on the given parameters

        This function performs a sensitivity analysis on the current
        experiment. It creates a new project and uses the evaluation and
        parameterization of the current experiment to get information on the
        others
        """
        from gwgen.sensitivity_analysis import SensitivityAnalysis
        sa_func_map = OrderedDict([
            ('setup', 'setup'), ('compile', 'compile_model'),
            ('init', 'init'), ('run', 'run'), ('evaluate', 'evaluate'),
            ('plot', 'plot'), ('remove', 'remove')])
        sensitivity_kws = OrderedDict(
            (key, kwargs[key]) for key in sa_func_map if key in kwargs)
        main_kws = {
            key: kwargs[key] for key in set(kwargs).difference(sa_func_map)}
        self.app_main(**main_kws)
        # to make sure, we already called the choose the right experiment and
        # projectname
        experiment = self.experiment
        self.logger.debug('Running sensitivity analysis for %s', experiment)
        sa = SensitivityAnalysis(self)
        self.fix_paths(self.exp_config)
        self.fix_paths(self.project_config)
        for key, val in sensitivity_kws.items():
            if isinstance(val, Namespace):
                val = vars(val)
            getattr(sa, sa_func_map[key])(**val)

    def _modify_sensitivity_analysis(self, parser):
        from gwgen.sensitivity_analysis import (
            SensitivityAnalysis, SensitivityPlot, default_sens_config)

        def params_type(s):
            splitted = s.split('=', 1)
            key = splitted[0]
            return key, utils.str_ranges(splitted[1])

        sps = parser.add_subparsers(help='Sensitivity analysis subroutines',
                                    chain=True)

        # setup parser
        sp = sps.add_parser('setup',
                            help='Setup the sensitivity analysis model')
        sp.setup_args(SensitivityAnalysis.setup)
        self._modify_app_main(sp)
        sp.update_arg('no_move', short='nm')

        # compile parser
        sp = sps.add_parser('compile',
                            help='Compile the sensitivity analysis model')
        sp.setup_args(SensitivityAnalysis.compile_model)
        self._modify_compile_model(sp)

        # init parser
        sp = sps.add_parser(
            'init', help='Initialize the sensitivity analysis experiments')
        sp.setup_args(SensitivityAnalysis.init)
        sp.update_arg('experiment', short='id')
        sp.update_arg(
            'nml', long='namelist', type=params_type,
            help=docstrings.dedents("""
                A list from namelist parameters and their values to use.
                %(str_ranges.s_help)s.
                You can also use ``'<i>err'`` in the list which will be
                interpreted as ``'<i>'``-times the error from the
                parameterization.
                """),
            metavar='nml_key=f1[,f21[-f22[-f23]]]', nargs='+')
        sp.update_arg('run_prepare', short='prep')
        sp.update_arg('no_move', short='nm')

        # run parser
        sp = sps.add_parser(
            'run', help='Run the sensitivity analysis experiments')
        sp.setup_args(SensitivityAnalysis.run)
        sp.update_arg('remove', short='rm')
        sp.update_arg('experiments', short='ids', type=lambda s: s.split(','),
                      metavar='id1,id2,...')
        sp.pop_key('experiments', 'nargs', None)

        # evaluate parser
        sp = sps.add_parser(
            'evaluate', help='Evaluate the sensitivity analysis experiments')
        sp.setup_args(SensitivityAnalysis.evaluate)
        sp.setup_args(self.evaluate)
        sp.update_arg('experiments', short='ids', type=lambda s: s.split(','),
                      metavar='id1,id2,...')
        sp.pop_key('experiments', 'nargs', None)
        sp.update_arg('loop_exps', short='loop')
        self._modify_evaluate(sp, skip=['prepare', 'output'])

        # plot parser
        sp = sps.add_parser(
            'plot', help='Plot the results sensitivity analysis experiments')
        sp.setup_args(SensitivityAnalysis.plot)
        defaults = default_sens_config()
        sp.update_arg('names', short='n', type=lambda s: s.split(','),
                      metavar='variable,[variable[,...]]',
                      default=defaults.names)
        sp.update_arg('indicators', short='i', type=lambda s: s.split(','),
                      metavar='indicator[,indicator[,...]]',
                      default=defaults.indicators)
        sp.update_arg('meta', metavar='<yaml-file>')
        tasks = utils.unique_everseen(
            SensitivityPlot.get_manager().sort_by_requirement(
                SensitivityPlot._registry[::-1]), lambda t: t.name)
        plot_sps = sp.add_subparsers(help='Plotting tasks', chain=True)
        for task in tasks:
            plot_sp = plot_sps.add_parser(task.name, help=task.summary)
            task._modify_parser(plot_sp)

        # remove parser
        sp = sps.add_parser('remove', help="Remove the sensitivity project")
        sp.setup_args(SensitivityAnalysis.remove)
        sp.setup_args(self.remove)
        self._modify_remove(sp)
        sp.pop_arg('projectname')
        sp.pop_arg('complete')

    # -------------------------------------------------------------------------
    # ------------------------------ Miscallaneous ----------------------------
    # -------------------------------------------------------------------------

    def _get_stations(self, stations, other_exp=False, odir=None,
                      config_key=None):
        """
        Get the stations for the parameterization or evaluation

        Parameters
        ----------
        stations: str or list of str
            either a list of stations to use or a filename containing a
            1-row table with stations
        other_exp: str
            Use the configuration from another experiment
        odir: str
            The output directory in case a list of stations is provided
        config_key:
            The key in the :attr:`exp_config` configuration dictionary holding
            information on the stations
        """
        import numpy as np

        exp_dict = self.exp_config
        fname = osp.join(odir, 'stations.dat') if odir else ''
        if other_exp and stations is None:
            stations = self.fix_paths(
                self.config.experiments[other_exp]).get(config_key)
        if isinstance(stations, six.string_types):
            stations = [stations]
        if stations is None:
            try:
                fname = exp_dict[config_key]
            except KeyError:
                raise ValueError('No stations file specified!')
            else:
                stations = np.loadtxt(exp_dict[config_key],
                                      dtype='S300', usecols=[0]).astype(
                    np.str_)
        elif len(stations) == 1 and osp.exists(stations[0]):
            fname_use = stations[0]
            exists = osp.exists(fname) if fname else False
            if exists and not osp.samefile(fname, fname_use):
                os.remove(fname)
                self._link(fname_use, fname)
            elif not exists and fname:
                self._link(fname_use, fname)
            stations = np.loadtxt(
                fname_use, dtype='S300', usecols=[0]).astype(np.str_)
        elif len(stations) and fname:
            np.savetxt(fname, stations, fmt='%s')
        if config_key and (not exp_dict.get(config_key) or not osp.samefile(
                fname, exp_dict[config_key])):
            exp_dict[config_key] = fname
        return stations

    def _setup_manager(
            self, manager, stations=None, other_exp=None,
            setup_from=None, to_db=None, to_csv=None, database=None,
            to_return=None, complete=False, base_kws={}):
        """
        Setup the data in a task manager

        This method is called by :meth:`param` and :meth:`evaluate` to setup
        the data in the given `manager`

        Parameters
        ----------
        manager: gwgen.utils.TaskManager
            The manager of the tasks to set up
        stations: list of str
            a list of stations to use
        other_exp: str
            Use the configuration from another experiment instead of
        setup_from: str
            Determine where to get the data from. If `scratch`, the
            data will be calculated from the raw data. If `file`,
            the data will be loaded from a file, if `db`, the data
            will be loaded from a postgres database (Note that the
            `database` argument must be provided!).
        to_db: bool
            Save the data into a postgresql database (Note that the
            `database` argument must be provided!)
        to_csv: bool
            Save the data into a csv file
        database: str
            The name of a postgres data base to write the data to
        to_return: list of str
            The names of the tasks to return. If None, only the ones with an
            :attr:`gwgen.utils.TaskBase.has_run` are returned.
        complete: bool
            If True, setup and run all possible tasks
        base_kws: dict
            The dictionary with mapping from each task name to the
            corresponding initialization keywords
        """
        if complete:
            for task in manager.base_task._registry:
                base_kws.setdefault(task.name, {})
        experiment = self.experiment
        exp_dict = self.fix_paths(self.config.experiments[experiment])
        if database is not None:
            exp_dict['database'] = database
        # setup up the keyword arguments for the parameterization tasks
        for key, d in base_kws.items():
            if d.get('setup_from') is None:
                d['setup_from'] = setup_from
            if to_csv:
                d['to_csv'] = to_csv
            elif to_csv is None and d.get('to_csv') is None:
                # delete the argument if the subparser doesn't use it
                d.pop('to_csv', None)
            if to_db:
                # delete the argument if the subparser doesn't use it
                d['to_db'] = to_db
            elif to_db is None and d.get('to_db') is None:
                d.pop('to_db', None)
            if other_exp and not d.get('other_exp'):
                d['other_exp'] = other_exp
            exp = d.pop('other_exp', experiment) or experiment
            d['config'] = self.fix_paths(self.config.experiments[exp])
            d['project_config'] = self.config.projects[d['config']['project']]
            for key in ['stations', 'complete', 'norun', 'other_id',
                        'database']:
                d.pop(key, None)
        # choose keywords for data processing
        manager.initialize_tasks(stations, task_kws=base_kws)
        manager.setup(stations, to_return=to_return)

    def _modify_task_parser(self, parser, base_task, skip=None, only=None):
        def norun(s):
            if s is True or s == 'all':
                return True
            try:
                return bool(int(s))
            except TypeError:
                return s.split(',')
        skip = skip or []
        if only is None:
            def key_func(t):
                return t.name not in skip
        else:
            def key_func(t):
                return t.name in only and t.name not in skip
        self._modify_app_main(parser)
        parser.update_arg('setup_from', short='f', long='from',
                          dest='setup_from')
        parser.update_arg('other_exp', short='ido', long='other_id',
                          dest='other_exp')
        try:
            parser.update_arg('stations', short='s')
        except KeyError:
            pass
        parser.update_arg('database', short='db')
        parser.pop_arg('to_return', None)
        parser.update_arg(
            'norun', short='nr', const=True, nargs='?',
            type=norun, help=(
                'If set without value or "all" or a number different from 0, '
                'the data is set up and the configuration of the '
                'experiment is not affected. Otherwise it can be a comma '
                'separated list of parameterization tasks for which to only '
                'setup the data'), metavar='task1,task2,...')
        doc = docstrings.params['GWGENOrganizer.param.parameters']
        setup_from_doc, setup_from_dtype = parser.get_param_doc(
            doc, 'setup_from')
        other_exp_doc, other_exp_dtype = parser.get_param_doc(doc, 'other_exp')

        tasks = filter(key_func, utils.unique_everseen(
            base_task.get_manager().sort_by_requirement(
                base_task._registry[::-1]), lambda t: t.name))
        sps = parser.add_subparsers(title='Tasks', chain=True)
        for task in tasks:
            sp = sps.add_parser(task.name, help=task.summary,
                                formatter_class=RawTextHelpFormatter)
            task._modify_parser(sp)
            sp.add_argument(
                '-ido', '--other_id', help=other_exp_doc,
                metavar=other_exp_dtype)

    def _link(self, source, target):
        """Link two files

        Parameters
        ----------
        source: str
            The path of the source file
        target: str
            The path of the target file"""
        if self.global_config.get('copy', True) and osp.isfile(source):
            shutil.copyfile(source, target)
        elif self.global_config.get('use_relative_links', True):
            os.symlink(osp.relpath(source, osp.dirname(target)), target)
        else:
            os.symlink(osp.abspath(source), target)


def exponential_function(x, a, b):
    """
    Exponential function used by :meth:`GWGENOrganizer.wind_bias_correction`

    This function is defined as

    .. math::

        f(x) = e^{ax + b}

    Parameters
    ----------
    x: numpy.ndarray
        The x-data
    a: float
        The *a* parameter in the above equation
    b: float
        The *b* parameter in the above equation

    Returns
    -------
    np.ndarray
        The calculated :math:`f(x)`
    """
    return np.exp(a * x + b)


def logistic_function(x, L, k, x0):
    """Logistic function used in :meth:`GWGENOrganizer.wind_bias_correction`

    The function is defined as

    .. math::

        f(x) = \\frac{L}{1 + \\mathrm e^{-k(x-x_0)}}

    Parameters
    ----------
    x: numpy.ndarray
        The x-data
    L: float
        the curve's maximum value
    k: float
        The steepness of the curve
    x0: the x-value of the sigmoid's midpoint

    Returns
    -------
    np.ndarray
        The calculated :math:`f(x)`"""
    return L / (1 + np.exp(-k * (x - x0)))


def _get_parser():
    """Function returning the gwgen parser, necessary for sphinx documentation
    """
    return GWGENOrganizer.get_parser()


def main(args=None):
    """Call the :meth:`~model_organization.GWGENOrganizer.main` method of the
    :class:`GWGENOrganizer` class"""
    GWGENOrganizer.main(args)


if __name__ == '__main__':
    main()



"""GWGEN: A WGEN-like weather generator for global daily weather
"""
from gwgen.version import __version__

__author__ = "Philipp Sommer and Jed Kaplan"
__maintainer__ = "Philipp Sommer"
__licence__ = "GPLv2"
__email__ = "philipp.sommer@unil.ch"
__copyright__ = "Copyright 2016, ARVE-Research"



# -*- coding: utf-8 -*-
"""Evaluation module of the gwgen module"""
from __future__ import division
import os.path as osp
import six
from collections import namedtuple
from psyplot.compat.pycompat import OrderedDict
from itertools import chain, starmap, repeat
import numpy as np
from scipy import stats
import pandas as pd
from gwgen.utils import docstrings
import gwgen.utils as utils
import logging


class Evaluator(utils.TaskBase):
    """Abstract base class for evaluation tasks

    Evaluation tasks should incorporate a run method that is called by the
    :meth:`gwgen.main.GWGENOrganizer.evaluate` method"""

    _registry = []

    @property
    def task_data_dir(self):
        """The directory where to store data"""
        return self.eval_dir

_PreparationConfigBase = namedtuple('_PreparationConfigBase',
                                    ['setup_raw', 'raw2db', 'raw2csv',
                                     'reference', 'input_path'])

_PreparationConfigBase = utils.append_doc(
    _PreparationConfigBase, docstrings.get_sections("""
Parameters
----------
setup_raw: { 'scratch' | 'file' | 'db' | None }
    The method how to setup the raw data from GHCN and EECRA

    ``'scratch'``
        To set up the task from the raw data
    ``'file'``
        Set up the task from an existing file
    ``'db'``
        Set up the task from a database
    ``None``
        If the file name of this this task exists, use this one,
        otherwise a database is provided, use this one, otherwise go
        from scratch
raw2db: bool
    If True, the raw data from GHCN and EECRA is stored in a postgres database
raw2csv: bool
    If True, the raw data from GHCN and EECRA is stored in a csv file
reference: str
    The path of the file where to store the reference data. If None and not
    already set in the configuration, it will default to
    ``'evaluation/reference.csv'``
input_path: str
    The path of the file where to store the model input. If None, and not
    already set in the configuration, it will default to
    ``'inputdir/input.csv'`` where *inputdir* is the path to the input
    directory (by default, *input* in the experiment directory)
""", '_PreparationConfigBase'))


PreparationConfig = utils.enhanced_config(_PreparationConfigBase,
                                          'PreparationConfig')


@docstrings.dedent
def default_preparation_config(
        setup_raw=None, raw2db=False, raw2csv=False, reference=None,
        input_path=None, *args, **kwargs):
    """
    The default configuration for :class:`EvaluationPreparation` instances.
    See also the :attr:`EvaluationPreparation.default_config` attribute

    Parameters
    ----------
    %(PreparationConfig.parameters)s"""
    return PreparationConfig(setup_raw, raw2db, raw2csv, reference, input_path,
                             *utils.default_config(*args, **kwargs))


class EvaluationPreparation(Evaluator):
    """Evaluation task to prepare the evaluation"""

    name = 'prepare'

    summary = 'Prepare the for experiment for evaluation'

    http_inventory = (
        'ftp://ftp.ncdc.noaa.gov/pub/data/ghcn/daily/ghcnd-inventory.txt')

    _datafile = ['reference.csv', 'input.csv']

    dbname = ['reference', 'input']

    has_run = False

    @property
    def default_config(self):
        ret = default_preparation_config()._replace(
            **super(EvaluationPreparation, self).default_config._asdict())
        return ret._replace(**dict(reference=self.reference_path,
                                   input_path=self.input_path))

    @property
    def datafile(self):
        """The paths to reference and input file"""
        return [self.reference_path, self.input_path]

    @property
    def ghcnd_inventory_file(self):
        return osp.join(self.data_dir, 'ghcn', 'ghcnd-inventory.txt')

    @property
    def station_list(self):
        fname = self.ghcnd_inventory_file
        station_ids, lat, lon, variables, first, last = np.loadtxt(
            fname, dtype='S11', unpack=True).astype(np.str_)
        return pd.DataFrame({'id': station_ids, 'lat': lat.astype(float),
                             'lon': lon.astype(float), 'vname': variables,
                             'firstyr': first.astype(int),
                             'lastyr': last.astype(int)},
                            index=np.arange(len(station_ids)))

    @property
    def reference_data(self):
        """The reference :class:`~pandas.DataFrame`"""
        return self.data[0]

    @reference_data.setter
    def reference_data(self, data):
        self.data[0] = data

    @property
    def input_data(self):
        """The input :class:`~pandas.DataFrame`"""
        return self.data[1]

    @input_data.setter
    def input_data(self, data):
        self.data[1] = data

    @classmethod
    def _modify_parser(cls, parser):
        parser.setup_args(default_preparation_config)
        # disable plot_output, etc.
        parser, setup_grp, run_grp = super(
            EvaluationPreparation, cls)._modify_parser(parser)
        parser.update_arg('setup_raw', short='fr', long='raw-from',
                          group=setup_grp)
        parser.update_arg('input_path', short='i', long='input',
                          group=setup_grp)
        parser.update_arg('reference', short='r', group=setup_grp)
        parser.update_arg('raw2db', group=setup_grp)
        parser.update_arg('raw2csv', group=setup_grp)
        return parser, setup_grp, run_grp

    def write2file(self, *args, **kwargs):
        """Reimplemented to sort the data according to the index"""
        for data in self.data:
            data.sort_index(inplace=True)
        return super(EvaluationPreparation, self).write2file(*args, **kwargs)

    def write2db(self, *args, **kwargs):
        """Reimplemented to sort the data according to the index"""
        for data in self.data:
            data.sort_index(inplace=True)
        return super(EvaluationPreparation, self).write2db(*args, **kwargs)

    def setup_from_db(self, *args, **kwargs):
        kwargs['index_col'] = [['id', 'year', 'month', 'day'],  # reference
                               ['id', 'lon', 'lat', 'year', 'month']]  # input
        super(EvaluationPreparation, self).setup_from_db(*args, **kwargs)

    def setup_from_file(self, *args, **kwargs):
        kwargs['index_col'] = [['id', 'year', 'month', 'day'],  # reference
                               ['id', 'lon', 'lat', 'year', 'month']]  # input
        super(EvaluationPreparation, self).setup_from_file(*args, **kwargs)

    def __init__(self, *args, **kwargs):
        self._manager = kwargs.pop('manager', None)
        super(EvaluationPreparation, self).__init__(*args, **kwargs)
        if self.task_config.to_csv:  # update the configuration
            self.reference_path = self.reference_path
            self.input_path = self.input_path

    def __reduce__(self):
        """Reimplemented to give provide also the manager"""
        ret = list(super(EvaluationPreparation, self).__reduce__())
        if len(ret) < 3:
            ret.append({})
        ret[2]['_manager'] = self._manager
        return tuple(ret)

    def init_from_scratch(self):
        """Initialize the setup via the parameterization classes"""
        from gwgen.parameterization import (
            Parameterizer, YearlyCompleteDailyCloud,
            YearlyCompleteDailyGHCNData, YearlyCompleteMonthlyCloud,
            YearlyCompleteMonthlyGHCNData)
        classes = [
            YearlyCompleteDailyCloud, YearlyCompleteDailyGHCNData,
            YearlyCompleteMonthlyCloud, YearlyCompleteMonthlyGHCNData]
        config = self.config.copy()
        config['paramdir'] = self.eval_dir
        kws = dict(config=config, project_config=self.project_config,
                   to_csv=self.task_config.raw2csv,
                   to_db=self.task_config.raw2db,
                   setup_from=self.task_config.setup_raw)
        base_kws = {cls.name: kws for cls in classes}
        self._manager = Parameterizer.get_manager(
            config=self.global_config.copy())
        self._manager.initialize_tasks(self.stations, task_kws=base_kws)
        if not self.global_config.get('serial'):
            import multiprocessing as mp
            for task in self._manager.tasks:
                utils._db_locks[task.dbname] = mp.Lock()
                utils._file_locks[task.datafile] = mp.Lock()
        # make sure the lat-lon information is available
        self.download_src()

    def download_src(self, force=False):
        fname = self.ghcnd_inventory_file
        if force or not osp.exists(fname):
            self.logger.info("Downloading station file from %s to %s",
                             self.http_inventory, fname)
            utils.download_file(self.http_inventory, fname)

    def setup_from_scratch(self):
        from gwgen.parameterization import (
            YearlyCompleteDailyCloud, YearlyCompleteDailyGHCNData,
            YearlyCompleteMonthlyCloud, YearlyCompleteMonthlyGHCNData)

        def get(name):
            return next(t for t in tasks if t.name == name)

        # force serial setup because we might already be in a parallel setup
        self._manager.config['serial'] = True
        self._manager.setup(self.stations, to_return='all')
        tasks = self._manager.tasks
        # save in the right order for the FORTRAN code
        order = ['tmin', 'tmax', 'mean_cloud', 'wind', 'prcp', 'wet_day']

        # daily reference
        cday = get(YearlyCompleteDailyGHCNData.name).data
        ccday = get(YearlyCompleteDailyCloud.name).data
        try:
            reference = cday.merge(
                ccday[['mean_cloud', 'wind']], left_index=True,
                right_index=True, how='left')
        except TypeError:  # indices to not match
            reference = cday.ix[1:0]  # create empty data frame
            reference['mean_cloud'] = np.array([],
                                               dtype=ccday.mean_cloud.dtype)
            reference['wind'] = np.array([],
                                         dtype=ccday.wind.dtype)
        reference['wet_day'] = (reference.prcp > 0).astype(int)
        self.reference_data = reference[order].sort_index()

        # monthly input
        cmonth = get(YearlyCompleteMonthlyGHCNData.name).data
        ccmonth = get(YearlyCompleteMonthlyCloud.name).data
        try:
            exp_input = cmonth.merge(
                ccmonth[['mean_cloud', 'wind']], left_index=True,
                right_index=True, how='left')
            # set cloud and wind to 0 where we have no reference
            exp_input.ix[exp_input.mean_cloud.isnull(),
                         ['mean_cloud', 'wind']] = 0
        except TypeError:  # indices do not match
            exp_input = cmonth.ix[1:0]  # create empty data frame
            exp_input['mean_cloud'] = np.array([],
                                               dtype=ccmonth.mean_cloud.dtype)
            exp_input['wind'] = np.array([],
                                         dtype=ccmonth.wind.dtype)
        inventory = self.station_list
        exp_input = exp_input.reset_index().merge(
            inventory[inventory.vname == 'PRCP'][['id', 'lon', 'lat']],
            on='id')
        exp_input.set_index(['id', 'lon', 'lat', 'year', 'month'],
                            inplace=True)
        self.input_data = exp_input[order].sort_index()


class OutputTask(Evaluator):
    """Task to provide all the data for input and output"""

    # the last will be ignored!
    _datafile = 'output.csv'

    dbname = 'output'

    name = 'output'

    summary = 'Load the output of the model'

    has_run = False

    @property
    def datafile(self):
        return [self.output_path]

    def write2file(self, *args, **kwargs):
        """Not implemented since the output file is generated by the model!"""
        self.logger.warn(
            "Writing to file of %s task is disabled because this is done "
            "by the model!", self.name)

    def setup_from_db(self, *args, **kwargs):
        kwargs['index_col'] = ['id', 'year', 'month', 'day']
        super(OutputTask, self).setup_from_db(*args, **kwargs)

    def setup_from_file(self, *args, **kwargs):
        kwargs['index_col'] = ['id', 'year', 'month', 'day']
        super(OutputTask, self).setup_from_file(*args, **kwargs)

    def setup_from_scratch(self):
        raise ValueError(
            "Cannot set up the %s task from scratch because it requires the "
            "output of the model!")


_KSConfig = namedtuple('_KSConfig', ['no_rounding', 'names', 'transform_wind'])


_KSConfig = utils.append_doc(_KSConfig, docstrings.get_sections("""
Parameters
----------
no_rounding: bool
    Do not round the simulation to the infered precision of the
    reference. The inferred precision is the minimum difference between
    two values with in the entire data
names: list of str
    The list of variables use for calculation. If None, all variables will be
    used
transform_wind: bool
    If True, the square root of the wind is evaluated (as this is also
    simulated in the weather generator)""", '_KSConfig'))


KSConfig = utils.enhanced_config(_KSConfig, 'KSConfig')


@docstrings.dedent
def default_ks_config(
        no_rounding=False, names=None, transform_wind=False, *args, **kwargs):
    """
    The default configuration for :class:`KSEvaluation` instances.
    See also the :attr:`KSEvaluation.default_config` attribute

    Parameters
    ----------
    %(KSConfig.parameters)s"""
    return KSConfig(no_rounding, names, transform_wind,
                    *utils.default_config(*args, **kwargs))


_QuantileConfig = namedtuple(
    '_QuantileConfig',
    ['quantiles'] + list(_KSConfig._fields))


_QuantileConfig = utils.append_doc(_QuantileConfig, docstrings.get_sections(
    docstrings.dedents("""
Parameters
----------
%(_KSConfig.parameters)s
quantiles: list of floats
    The quantiles to use for calculating the percentiles
"""), '_QuantileConfig'))


QuantileConfig = utils.enhanced_config(_QuantileConfig, 'QuantileConfig')


@docstrings.dedent
def default_quantile_config(
        quantiles=[1, 5, 10, 25, 50, 75, 90, 95, 99, 100], *args, **kwargs):
    """
    The default configuration for :class:`QuantileEvaluation` instances.
    See also the :attr:`QuantileEvaluation.default_config` attribute

    Parameters
    ----------
    %(QuantileConfig.parameters)s"""
    return QuantileConfig(quantiles, *default_ks_config(*args, **kwargs))


class QuantileEvaluation(Evaluator):
    """Evaluator to evaluate specific quantiles"""

    name = 'quants'

    summary = 'Compare the quantiles of simulation and observation'

    names = OrderedDict([
        ('prcp', {'long_name': 'Precipitation',
                  'units': 'mm'}),
        ('tmin', {'long_name': 'Min. Temperature',
                  'units': 'degC'}),
        ('tmax', {'long_name': 'Max. Temperature',
                  'units': 'degC'}),
        ('mean_cloud', {'long_name': 'Cloud fraction',
                        'units': '-'}),
        ('wind', {'long_name': 'Wind Speed',
                  'units': 'm/s'})
        ])

    @property
    def all_variables(self):
        return [[v + '_ref', v + '_sim'] for v in self.names]

    setup_requires = ['prepare', 'output']

    has_run = True

    _datafile = 'quantile_evaluation.csv'

    dbname = 'quantile_evaluation'

    #: default formatoptions for the
    #: :class:`psyplot.plotter.linreg.DensityRegPlotter` plotter
    fmt = kwargs = dict(
        legend={'loc': 'upper left'},
        cmap='w_Reds',
        title='%(pctl)sth percentile',
        xlabel='%(type)s {desc}',
        ylabel='%(type)s {desc}',
        xrange=(['minmax', 1], ['minmax', 99]),
        yrange=(['minmax', 1], ['minmax', 99]),
        legendlabels=['$R^2$ = %(rsquared)s'],
        bounds=['minmax', 11, 0, 99],
        cbar='',
        bins=10,
        ideal=[0, 1],
        id_color='r',
        sym_lims='max',
        )

    @property
    def default_config(self):
        return default_quantile_config()._replace(
            **super(QuantileEvaluation, self).default_config._asdict())

    @property
    def ds(self):
        """The dataset of the quantiles"""
        import xarray as xr
        data = self.data.reset_index()
        ds = xr.Dataset.from_dataframe(
            data.set_index('pctl', append=True).swaplevel())
        full = xr.Dataset.from_dataframe(data).drop(list(chain(
            self.data.index.names, ['pctl'])))
        idx_name = full[next(v for v in self.names) + '_sim'].dims[-1]
        full.rename({idx_name: 'full_index'}, inplace=True)
        for vref, vsim in self.all_variables:
            full.rename({vref: 'all_' + vref, vsim: 'all_' + vsim},
                        inplace=True)
        ds.merge(full, inplace=True)
        for orig, attrs, (vref, vsim) in zip(
                self.names, self.names.values(), self.all_variables):
            for prefix in ['', 'all_']:
                ds[prefix + vsim].attrs.update(attrs)
                ds[prefix + vref].attrs.update(attrs)
                ds[prefix + vsim].attrs['standard_name'] = orig
                ds[prefix + vref].attrs['standard_name'] = orig
                ds[prefix + vref].attrs['type'] = 'observed'
                ds[prefix + vsim].attrs['type'] = 'simulated'
            ds['all_' + vsim].attrs['pctl'] = 'All'
            ds['all_' + vsim].attrs['pctl'] = 'All'
        ds.pctl.attrs['long_name'] = 'Percentile'
        return ds

    def __init__(self, *args, **kwargs):
        super(QuantileEvaluation, self).__init__(*args, **kwargs)
        names = self.task_config.names
        if names is not None:
            self.names = OrderedDict(
                t for t in self.names.items() if t[0] in names)

    def setup_from_file(self, *args, **kwargs):
        kwargs['index_col'] = ['id', 'year']
        return super(QuantileEvaluation, self).setup_from_file(*args, **kwargs)

    def setup_from_db(self, *args, **kwargs):
        kwargs['index_col'] = ['id', 'year']
        return super(QuantileEvaluation, self).setup_from_db(*args, **kwargs)

    def setup_from_scratch(self):
        df_ref = self.prepare.reference_data
        dfi = self.prepare.input_data.reset_index(['lon', 'lat'])
        # create simulation dataframe
        df_sim = self.output.data
        if len(df_ref) == 0 or len(df_sim) == 0:
            self.logger.debug(
                'Skipping %s because reference data contains no information!',
                self.name)
            return
        names = self.names
        # load observed precision
        if self.task_config.no_rounding:
            for name in names:
                df_sim[name].values[:] = self.round_to_ref_prec(
                    df_ref[name].values, df_sim[name].values)
        # merge reference and simulation into one single dataframe
        df = df_ref.merge(df_sim, left_index=True, right_index=True,
                          suffixes=['_ref', '_sim'])
        if {'mean_cloud', 'wind'}.intersection(names):
            df.reset_index('day', inplace=True)
            df = df.merge(dfi[['mean_cloud', 'wind']], left_index=True,
                          right_index=True)
            # mask out non-complete months for cloud validation and months with
            # 0 or 1 cloud fraction
            if 'mean_cloud' in names:
                df.ix[df['mean_cloud_ref'].isnull().values |
                      (df['mean_cloud'] == 0.0) |
                      (df['mean_cloud'] == 1.0),
                      ['mean_cloud_sim', 'mean_cloud_ref']] = np.nan
            # mask out non-complete wind for wind validation and months with
            # a mean wind speed of 0
            if 'wind' in names:
                df.ix[df['wind_ref'].isnull().values | (df['wind'] == 0.0),
                      ['wind_sim', 'wind_ref']] = np.nan
            df.drop(['mean_cloud', 'wind'], 1, inplace=True)
            df.set_index('day', append=True, inplace=True)

        # transform wind
        if self.task_config.transform_wind and 'wind' in names:
            df['wind_ref'] **= 0.5
            df['wind_sim'] **= 0.5
        # calculate the percentiles for each station and month
        g = df.sort_index().groupby(level=['id', 'year'])
        self.logger.debug('Done with basic setup')
        data = g.apply(self.calc)
        if len(data):
            data.index = data.index.droplevel(2)
        self.data = data

    @classmethod
    def _modify_parser(cls, parser):
        parser.setup_args(default_ks_config)
        parser.setup_args(default_quantile_config)
        parser, setup_grp, run_grp = super(
            QuantileEvaluation, cls)._modify_parser(parser)
        parser.update_arg(
            'quantiles', short='q', group=run_grp, type=utils.str_ranges,
            metavar='f1[,f21[-f22[-f23]]]', help=docstrings.dedents("""
                The quantiles to use for calculating the percentiles.
                %(str_ranges.s_help)s."""))
        parser.pop_key('quantiles', 'nargs', None)
        parser.update_arg('no_rounding', short='nr', group=run_grp)
        parser.update_arg('names', short='n', group=setup_grp,
                          nargs='+', metavar='variable',
                          choices=list(cls.names))
        parser.update_arg('transform_wind', short='tw', group=setup_grp)
        return parser, setup_grp, run_grp

    def create_project(self, ds):
        import psyplot.project as psy
        import seaborn as sns
        sns.set_style('white')
        for name, (vref, vsim) in zip(self.names, self.all_variables):
            self.logger.debug('Creating plots of %s', vsim)
            kwargs = dict(precision=0.1) if vref.startswith('prcp') else {}
            psy.plot.densityreg(ds, name='all_' + vsim, coord='all_' + vref,
                                fmt=self.fmt, title='All percentiles',
                                arr_names=['%s_all' % name],
                                **kwargs)
            arr_names = ['%s_%1.2f' % (name, p) for p in ds.pctl.values]
            psy.plot.densityreg(ds, name=vsim, coord=vref, fmt=self.fmt,
                                arr_names=arr_names, pctl=range(ds.pctl.size),
                                **kwargs)
        return psy.gcp(True)[:]

    def make_run_config(self, sp, info):
        for orig in self.names:
            info[orig] = d = OrderedDict()
            for plotter in sp(standard_name=orig).plotters:
                d[plotter.data.pctl if plotter.data.name.startswith('all') else
                  int(plotter.data.pctl.values)] = pctl_d = OrderedDict()
                for key in ['rsquared', 'slope', 'intercept']:
                        val = plotter.plot_data[1].attrs.get(key)
                        if val is not None:
                            pctl_d[key] = float(val)
        return info

    def calc(self, group):
        def calc_percentiles(vname):
            arr = group[vname].values
            arr = arr[~np.isnan(arr)]
            if vname.startswith('prcp'):
                arr = arr[arr > 0]
            if len(arr) == 0:
                return np.array([np.nan] * len(quantiles))
            else:
                return np.percentile(arr, quantiles)
        quantiles = self.task_config.quantiles
        df = pd.DataFrame.from_dict(dict(zip(
            chain(*self.all_variables), map(calc_percentiles,
                                            chain(*self.all_variables)))))
        df['pctl'] = quantiles
        df.set_index('pctl')
        return df

    @staticmethod
    def round_to_ref_prec(ref, sim, func=np.ceil):
        """Round one array to the precision of another

        Parameters
        ----------
        ref: np.ndarray
            The reference array to get the precision from
        sim: np.ndarray
            The simulated array to round
        func: function
            The rounding function to use

        Returns
        -------
        np.ndarray
            Rounded `sim`"""
        ref_sorted = np.unique(ref)
        if len(ref_sorted) < 2:
            return sim
        precision = (ref_sorted[1:] - ref_sorted[:-1]).min()
        return func((sim / precision) * precision)


class KSEvaluation(QuantileEvaluation):
    """Evaluation using a Kolmogorov-Smirnoff test"""

    name = 'ks'

    summary = 'Perform a kolmogorov smirnoff test'

    requires = ['prepare', 'output']

    _datafile = 'kolmogorov_evaluation.csv'

    dbname = 'kolmogorov_evaluation'

    @property
    def default_config(self):
        return default_ks_config()._replace(
            **super(QuantileEvaluation, self).default_config._asdict())

    @staticmethod
    def calc(group):
        def calc(v1, v2, name):
            if len(v1) <= 10 or len(v2) <= 10:
                return {
                    name + '_stat': [np.nan],
                    name + '_p': [np.nan],
                    name: [None],
                    'n' + name + '_sim': [np.nan],
                    'n' + name + '_ref': [np.nan]}
            statistic, p_value = stats.ks_2samp(v1, v2)
            n1 = len(v1)
            n2 = len(v2)
            n = np.sqrt((n1 + n2) / (n1 * n2))
            # if statistic > 1.36 * n, we reject the null hypothesis
            # (alpha = 0.05)
            return {
                name + '_stat': [statistic],
                name + '_p': [p_value],
                name: [statistic > 1.36 * n],
                'n' + name + '_sim': [n1],
                'n' + name + '_ref': [n2]}
        prcp_sim = group.prcp_sim.values[group.prcp_sim.values > 0]
        prcp_ref = group.prcp_ref.values[group.prcp_ref.values > 0]
        tmin_sim = group.tmin_sim.values[
            (group.tmin_sim.values < 100) & (group.tmin_sim.values > -100) &
            (~np.isnan(group.tmin_sim.values))]
        tmin_ref = group.tmin_ref.values[
            (group.tmin_ref.values < 100) & (group.tmin_ref.values > -100) &
            (~np.isnan(group.tmin_ref.values))]
        tmax_sim = group.tmax_sim.values[
            (group.tmax_sim.values < 100) & (group.tmax_sim.values > -100) &
            (~np.isnan(group.tmax_sim.values))]
        tmax_ref = group.tmax_ref.values[
            (group.tmax_ref.values < 100) & (group.tmax_ref.values > -100) &
            (~np.isnan(group.tmax_ref.values))]
        cloud_sl = group.mean_cloud_ref.notnull().values
        cloud_sim = group.mean_cloud_sim.values[cloud_sl]
        cloud_ref = group.mean_cloud_ref.values[cloud_sl]
        wind_sl = group.wind_ref.notnull().values
        wind_sim = group.wind_sim.values[wind_sl]
        wind_ref = group.wind_ref.values[wind_sl]
        return pd.DataFrame.from_dict(dict(chain(*map(six.iteritems, starmap(
            calc, [(prcp_sim, prcp_ref, 'prcp'),
                   (tmin_sim, tmin_ref, 'tmin'),
                   (tmax_sim, tmax_ref, 'tmax'),
                   (cloud_sim, cloud_ref, 'mean_cloud'),
                   (wind_sim, wind_ref, 'wind')])))))

    def significance_fractions(self, series):
        "The percentage of stations with no significant difference"
        return 100. - (len(series.ix[series.notnull() & (series)]) /
                       series.count())*100.

    def run(self, info):
        """Run the evaluation

        Parameters
        ----------
        info: dict
            The configuration dictionary"""
        logger = self.logger
        logger.info('Calculating %s evaluation', self.name)

        df = self.data
        names = list(self.names)
        for name in names:
            info[name] = float(self.significance_fractions(df[name]))

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('Done. Stations without significant difference:')
            for name, val in info.items():
                logger.debug('    %s: %6.3f %%' % (name, val))
        try:
            import cartopy.crs as ccrs
        except ImportError:
            self.logger.warn(
                "Cartopy is not installed, skipping plot of %s task",
                self.name)
        else:
            self.plot_map()
            info['plot_file'] = self.pdf_file

    def plot_map(self):
        from matplotlib.backends.backend_pdf import PdfPages
        import cartopy.crs as ccrs
        import matplotlib.pyplot as plt
        names = list(self.names)
        df = self.data[names]
        for n in names:
            df[n] = df.pop(n).astype(float)
        g = df.groupby(level='id')
        df_fract = g.agg(dict(zip(
            names, repeat(self.significance_fractions))))
        df_fract = df_fract.merge(
            g.agg(dict(zip(names, repeat('sum')))), left_index=True,
            right_index=True, suffixes=['', '_sum'])
        df_lola = EvaluationPreparation.from_task(self).station_list
        df_lola = df_lola.ix[~df_lola.duplicated('id').values]
        df_lola.set_index('id', inplace=True)
        df_plot = df_lola.merge(df_fract, how='right', left_index=True,
                                right_index=True)
        pdf = PdfPages(self.pdf_file)
        for name in names:
            fig = plt.figure()
            ax = plt.axes(projection=ccrs.PlateCarree())
            ax.set_title(self.names[name]['long_name'])
            ax.coastlines()
            ax.background_patch.set_facecolor('0.95')
            ax.set_global()
            df_plot.sort_values(name, ascending=False).plot.scatter(
                'lon', 'lat', c=name, ax=ax, colorbar=False,
                transform=ccrs.PlateCarree(), cmap='Reds_r', s=5,
                edgecolor='none')

            cbar = fig.colorbar(ax.collections[-1], orientation='horizontal')
            cbar.set_label(
                'Percentage of years not differing significantly[%]')

            pdf.savefig(fig, bbox_inches='tight')

            fig = plt.figure()
            ax = plt.axes(projection=ccrs.PlateCarree())
            ax.set_title(self.names[name]['long_name'])
            ax.coastlines()
            ax.background_patch.set_facecolor('0.95')
            ax.set_global()
            df_plot.sort_values(name, ascending=False).plot.scatter(
                'lon', 'lat', c=name + '_sum', ax=ax, colorbar=False,
                transform=ccrs.PlateCarree(), cmap='Reds', s=5,
                edgecolor='none')

            cbar = fig.colorbar(ax.collections[-1], orientation='horizontal')
            cbar.set_label('Significantly differing years')

            pdf.savefig(fig, bbox_inches='tight')
        pdf.close()
        if self.task_config.close:
            plt.close('all')

    @classmethod
    def _modify_parser(cls, parser):
        parser.setup_args(default_ks_config)
        parser, setup_grp, run_grp = super(
            QuantileEvaluation, cls)._modify_parser(parser)
        parser.update_arg('no_rounding', short='nr', group=run_grp)
        parser.update_arg('names', short='n', group=setup_grp,
                          nargs='+', metavar='variable',
                          choices=list(cls.names))
        parser.update_arg('transform_wind', short='tw', group=setup_grp)
        parser.pop_arg('nc_output', None)
        parser.pop_arg('project_output', None)
        parser.pop_arg('new_project', None)
        parser.pop_arg('project', None)
        return parser, setup_grp, run_grp


_QualityConfig = namedtuple('_QualityConfig', ['quantiles'])


_QualityConfig = utils.append_doc(_QualityConfig, docstrings.get_sections("""
Parameters
----------
quantiles: list of floats
    The quantiles to use for the quality analysis""", '_QualityConfig'))


QualityConfig = utils.enhanced_config(_QualityConfig, 'QualityConfig')


@docstrings.dedent
def default_quality_config(
        quantiles=None, *args, **kwargs):
    """
    The default configuration for :class:`SimulationQuality` instances.
    See also the :attr:`SimulationQuality.default_config` attribute

    Parameters
    ----------
    %(QualityConfig.parameters)s"""
    return QualityConfig(quantiles, *utils.default_config(*args, **kwargs))


class SimulationQuality(Evaluator):
    """Evaluator to provide one value characterizing the quality of the
    experiment

    The applied metric is the mean of

    .. math::
        m = \\left<\{\\left<\{R^2_q\}_{q\in Q}\\right>,
                     \\left<\{1 - |1 - a_q|\}_{q\in Q}\\right>,
                     \{ks\}\}\\right>

    where :math:`\\left<\\right>` denotes the mean of the enclosed set,
    :math:`q\in Q` are the quantiles from the quantile evaluation,
    :math:`R^2_q` the corresponding coefficient of determination and
    :math:`a_q` the slope of quantile :math:`q`. :math:`ks` denotes the
    fraction of stations that do not differ significantly from the observations
    according to the ks test.

    In other words, this quality estimate is the mean of the

        1. coefficients of determination
        2. the deviation from the ideal slope (:math:`a_q == 1`) and
        3. the fraction of stations that do not differ significantly

    Hence, a value of 1 mean high quality, a value of 0 low quality"""

    name = 'quality'

    summary = 'Estimate simulation quality using ks and quantile evaluation'

    has_run = True

    @classmethod
    def _modify_parser(cls, parser):
        parser.add_argument(
            '-q', '--quantiles', metavar='q1[,q1[,q2[,...]]]',
            help="The quantiles to use for the quality analysis",
            type=lambda s: list(map(float, s.split(','))))
        return parser, None, None

    @property
    def default_config(self):
        return default_quality_config()._replace(
            **super(SimulationQuality, self).default_config._asdict())

    def setup_from_scratch(self):
        """Only sets an empty dataframe"""
        self.data = pd.DataFrame([])

    def run(self, info):
        logger = self.logger
        logger.info('Calculating %s evaluation', self.name)
        quants_info = self.config['evaluation'].get('quants')
        ks_info = self.config['evaluation'].get('ks')
        missing = []
        if quants_info is None:
            missing.append('quants task')
        if ks_info is None:
            missing.append('ks task')
        if missing:
            raise ValueError("%s requires that %s has been run before!" % (
                self.name, ' and '.join(missing)))
        quantiles = self.task_config.quantiles
        if quantiles is None:
            quantiles = slice(None)
        possible_names = {'wind', 'prcp', 'tmin', 'tmax', 'mean_cloud',
                          'cloud'}
        for v, v_ks in ks_info.items():
            if v not in quants_info or v not in possible_names:
                continue
            #: Dataframe with intercept, rsquared and slope on index and
            #: quantiles as columns
            df = pd.DataFrame(quants_info[v]).loc[:, quantiles]
            try:
                del df['All']
            except KeyError:
                pass
            slope = float((1 - np.abs(1 - df.loc['slope'].values)).mean())
            rsquared = float(df.loc['rsquared'].values.mean())
            info[v] = OrderedDict([
                ('rsquared', rsquared), ('slope', slope), ('ks', v_ks / 100.),
                ('quality', float(np.mean([rsquared, slope, v_ks / 100.])))])
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('Simulation quality:')
            for name, val in info.items():
                logger.debug('    %s: %6.3f' % (name, val['quality']))



# -*- coding: utf-8 -*-
from gwgen._parseghcnrow import parseghcnrow
import six
import pandas as pd
import numpy as np
from itertools import chain
import datetime as dt
import re

if six.PY2:
    from itertools import imap as map, izip as zip

daymon_patt = re.compile(r'(?:\w|-){11}(\d{6})(?:TMAX|TMIN|PRCP)')


def read_ghcn_file(ifile):
    """Read in a GHCN station data file and convert it to a dataframe

    Parameters
    ----------
    ifile: str
        The path to a ghcn datafile

    Returns
    -------
    pandas.DataFrame
        The `ifile` converted to a dataframe"""
    # get number of days in the file
    with open(ifile) as f:
        ndays = np.sum(list(map(ndaymon, np.unique(list(map(
            lambda m: m.group(1), filter(
                None, map(daymon_patt.match, f.readlines()))))))))
    stationid, dates, variables, flags, j = parseghcnrow.parse_station(
        ifile, ndays or 100)
    dates = dates[:j]
    variables = variables[:j].astype(np.float64)
    flags = flags[:j].astype(np.str_)
    flags = np.core.defchararray.replace(flags, ' ', '')
    variables[np.isclose(variables, -9999.) |
              np.isclose(variables, -999.9)] = np.nan
    vlst = ['tmin', 'tmax', 'prcp']
    df = pd.DataFrame.from_dict(dict(chain(
        [('id', np.repeat(np.array([stationid]).astype(np.str_), j))],
        zip(('year', 'month', 'day'), dates.T),
        zip(vlst, variables.T),
        chain(*[zip((var + '_m', var + '_q', var + '_s'), arr)
                for var, arr in zip(vlst, np.rollaxis(flags, 2, 1).T)]))))
    return df


def ndaymon(yearmon):
    """Calculate the number of days for one month in a year

    Parameters
    ----------
    yearmon: str
        The first 4 numbers stand for the year, the others for the month
        (in datetime writing: ``'%Y%m'``)"""
    year = int(yearmon[:4])
    month = int(yearmon[4:])
    d = dt.date(year, month, 1)
    d2 = d.replace(
        year=year + 1 if month == 12 else year,
        month=1 if month == 12 else month + 1)
    return (d2 - d).days
